from __future__ import annotations

import linecache
from abc import ABC, abstractmethod
from collections import Counter
from typing import TYPE_CHECKING

import numpy as np
from monty.io import zopen
from monty.json import MSONable

from pymatgen.core import Lattice, Structure
from pymatgen.symmetry.kpath import KPathSeek

if TYPE_CHECKING:
    from typing import Self

    from pymatgen.core.structure import IStructure
    from pymatgen.util.typing import PathLike

__author__ = "Hanyu Liu"
__email__ = "domainofbuaa@gmail.com"
__date__ = "2024-1-16"


class LineLocator(MSONable):
    """Find the line indices (starts from 1) of a certain paragraph of text from the file."""

    @staticmethod
    def locate_all_lines(file_path: PathLike, content: str, exclusion: str = "") -> list[int]:
        """Locate the line in file where a certain paragraph of text is located (return all indices).

        Args:
            file_path (PathLike): Absolute path to file.
            content (str): Certain paragraph of text that needs to be located.
            exclusion (str): Certain paragraph of text that is excluded.
        """
        row_idxs: list[int] = []  # starts from 1 to be compatible with linecache package
        row_no: int = 0
        with zopen(file_path, mode="rt", encoding="utf-8") as file:
            for row_content in file:
                row_no += 1
                if content.upper() in row_content.upper() and (
                    not exclusion or exclusion.upper() not in row_content.upper()
                ):
                    row_idxs.append(row_no)
        return row_idxs


class ListLocator(MSONable):
    """Find the element indices (starts from 0) of a certain paragraph of text from the list."""

    @staticmethod
    def locate_all_lines(strs_lst: list[str], content: str, exclusion: str = "") -> list[int]:
        """Locate the elements in list where a certain paragraph of text is located (return all indices).

        Args:
            strs_lst (list[str]): List of strings.
            content (str): Certain paragraph of text that needs to be located.
            exclusion (str): Certain paragraph of text that is excluded.
        """
        str_idxs: list[int] = []  # starts from 0 to be compatible with list
        str_no: int = -1
        for tmp_str in strs_lst:
            str_no += 1
            if (content.upper() in tmp_str.upper()) and (not exclusion or exclusion.upper() not in tmp_str.upper()):
                str_idxs.append(str_no)
        return str_idxs


class ACExtractorBase(ABC):
    """A parent class of ACExtractor and ACstrExtractor, ensuring that they are as consistent as possible."""

    @abstractmethod
    def get_n_atoms(self) -> int:
        """Get the number of atoms in structure defined by atom.config file."""

    @abstractmethod
    def get_lattice(self) -> np.ndarray:
        """Get the lattice of structure defined by atom.config file."""

    @abstractmethod
    def get_types(self) -> np.ndarray:
        """Get atomic number of atoms in structure defined by atom.config file."""

    @abstractmethod
    def get_coords(self) -> np.ndarray:
        """Get fractional coordinates of atoms in structure defined by atom.config file."""

    @abstractmethod
    def get_magmoms(self) -> np.ndarray:
        """Get atomic magmoms of atoms in structure defined by atom.config file."""


class ACExtractor(ACExtractorBase):
    """Extract information contained in atom.config : number of atoms, lattice, types, frac_coords, magmoms."""

    def __init__(self, file_path: PathLike) -> None:
        """Initialization function.

        Args:
            file_path (str): The absolute path of atom.config file.
        """
        self.atom_config_path = file_path
        self.n_atoms = self.get_n_atoms()
        self.lattice = self.get_lattice()
        self.types = self.get_types()
        self.coords = self.get_coords()
        self.magmoms = self.get_magmoms()

    def get_n_atoms(self) -> int:
        """Return the number of atoms in the structure."""
        first_row = linecache.getline(str(self.atom_config_path), 1)
        return int(first_row.split()[0])

    def get_lattice(self) -> np.ndarray:
        """Return the lattice of structure.

        Returns:
            lattice: np.ndarray, shape = (9,)
        """
        basis_vectors: list[float] = []
        content: str = "LATTICE"
        idx_row: int = LineLocator.locate_all_lines(file_path=self.atom_config_path, content=content)[0]
        for row_idx in [idx_row + 1, idx_row + 2, idx_row + 3]:
            row_content: list[str] = linecache.getline(str(self.atom_config_path), row_idx).split()[:3]
            basis_vectors.extend(float(value) for value in row_content)

        return np.array(basis_vectors)

    def get_types(self) -> np.ndarray:
        """Return the atomic number of atoms in structure.

        Returns:
            np.ndarray: Atomic numbers in order corresponding to sites
        """
        content = "POSITION"
        idx_row = LineLocator.locate_all_lines(file_path=self.atom_config_path, content=content)[0]
        with open(self.atom_config_path, encoding="utf-8") as file:
            atom_config_content = file.readlines()
        atomic_numbers_content = atom_config_content[idx_row : idx_row + self.n_atoms]
        atomic_numbers_lst = [int(row.split()[0]) for row in atomic_numbers_content]  # convert str to int
        return np.array(atomic_numbers_lst)

    def get_coords(self) -> np.ndarray:
        """Return the fractional coordinates in structure.

        Returns:
            np.ndarray: Fractional coordinates.
        """
        coords_lst: list[np.ndarray] = []
        content: str = "POSITION"
        idx_row: int = LineLocator.locate_all_lines(file_path=self.atom_config_path, content=content)[0]
        with open(self.atom_config_path, encoding="utf-8") as file:
            atom_config_content = file.readlines()
        """
        row_content:
            '29         0.377262291145329         0.128590184800933         0.257759805813488     1  1  1'
        """
        for row_content in atom_config_content[idx_row : idx_row + self.n_atoms]:
            row_content_lst = row_content.split()
            coord_tmp = [float(value) for value in row_content_lst[1:4]]  # convert str to float.
            coords_lst.append(np.array(coord_tmp))
        return np.array(coords_lst).reshape(-1)

    def get_magmoms(self) -> np.ndarray:
        """Return the magenetic moments of atoms in structure.

        Returns:
            np.ndarray: The magnetic moments of individual atoms.
        """
        content: str = "MAGNETIC"
        magnetic_moments_lst: list[float] = []
        try:  # Error: not containing magmoms info.
            idx_row = LineLocator.locate_all_lines(file_path=self.atom_config_path, content=content)[-1]

            with open(self.atom_config_path, encoding="utf-8") as file:
                atom_config_content = file.readlines()

            magnetic_moments_content = atom_config_content[idx_row : idx_row + self.n_atoms]
            # MAGNETIC
            # 3 0.0 # atomic_number magmom
            # ...
            magnetic_moments_lst = [
                float(tmp_magnetic_moment.split()[-1]) for tmp_magnetic_moment in magnetic_moments_content
            ]
        except Exception:
            magnetic_moments_lst = [0 for _ in range(self.n_atoms)]
        return np.array(magnetic_moments_lst)


class ACstrExtractor(ACExtractorBase):
    """Extract information from atom.config file. You can get str by slicing the MOVEMENT."""

    def __init__(self, atom_config_str: str):
        """Initialization function.

        Args:
            atom_config_str (str): A string describing the structure in atom.config file.
        """
        self.atom_config_str = atom_config_str
        self.strs_lst = self.atom_config_str.split("\n")
        self.num_atoms = self.get_n_atoms()

    def get_n_atoms(self) -> int:
        """Return the number of atoms in structure.

        Returns:
            int: The number of atoms
        """
        return int(self.strs_lst[0].split()[0].strip())

    def get_lattice(self) -> np.ndarray:
        """Return the lattice of structure.

        Returns:
            np.ndarray: Lattice basis vectors of shape=(9,)
        """
        basis_vectors_lst: list[float] = []
        aim_content = "LATTICE"
        aim_idx = ListLocator.locate_all_lines(strs_lst=self.strs_lst, content=aim_content)[0]
        for idx_str in [aim_idx + 1, aim_idx + 2, aim_idx + 3]:
            # ['0.8759519000E+01', '0.0000000000E+00', '0.0000000000E+00']
            str_lst = self.strs_lst[idx_str].split()[:3]
            basis_vectors_lst.extend(float(tmp_str) for tmp_str in str_lst)
        return np.array(basis_vectors_lst)

    def get_types(self) -> np.ndarray:
        """Return the atomic number of atoms in structure.

        Returns:
            np.ndarray: Types of elements.
        """
        aim_content = "POSITION"
        aim_idx = ListLocator.locate_all_lines(strs_lst=self.strs_lst, content=aim_content)[0]
        strs_lst = self.strs_lst[aim_idx + 1 : aim_idx + self.num_atoms + 1]
        atomic_numbers_lst = [int(entry.split()[0]) for entry in strs_lst]
        return np.array(atomic_numbers_lst)

    def get_coords(self) -> np.ndarray:
        """Return the fractional coordinate of atoms in structure.

        Returns:
            np.ndarray: Fractional coordinates of atoms of shape=(num_atoms*3,)
        """
        coords_lst = []
        aim_content = "POSITION"
        aim_idx = ListLocator.locate_all_lines(strs_lst=self.strs_lst, content=aim_content)[0]
        for tmp_str in self.strs_lst[aim_idx + 1 : aim_idx + self.num_atoms + 1]:
            # ['14', '0.751401861790384', '0.501653718883189', '0.938307102003243', '1', '1', '1']
            tmp_strs_lst = tmp_str.split()
            tmp_coord = [float(value) for value in tmp_strs_lst[1:4]]  # convert str to float
            coords_lst.append(tmp_coord)
        return np.array(coords_lst).reshape(-1)

    def get_magmoms(self) -> np.ndarray:
        """Return the magnetic moments of atoms in structure.

        Returns:
            np.ndarray: Atomic magnetic moments.
        """
        magnetic_moments_lst: list[float] = []
        aim_content: str = "MAGNETIC"
        aim_idxs: list[int] = ListLocator.locate_all_lines(strs_lst=self.strs_lst, content=aim_content)
        if len(aim_idxs) == 0:
            magnetic_moments_lst = [0.0 for _ in range(self.num_atoms)]
        else:
            aim_idx = aim_idxs[0]
            magnetic_moments_content = self.strs_lst[aim_idx + 1 : aim_idx + self.num_atoms + 1]
            magnetic_moments_lst = [
                float(tmp_magnetic_moment.split()[-1]) for tmp_magnetic_moment in magnetic_moments_content
            ]
        return np.array(magnetic_moments_lst)

    def get_e_tot(self) -> np.ndarray:
        """Return the total energy of structure.

        Returns:
            np.ndarray: The total energy of the material system.
        """
        # strs_lst:
        #   [' 216 atoms', 'Iteration (fs) =    0.3000000000E+01',
        #    ' Etot', 'Ep', 'Ek =   -0.2831881714E+05  -0.2836665392E+05   0.4783678177E+02',
        #    ' SCF =     7']
        strs_lst = self.strs_lst[0].split(",")
        aim_index = ListLocator.locate_all_lines(strs_lst=strs_lst, content="EK")[0]
        # strs_lst[aim_index].split() :
        #   ['Ek', '(eV)', '=', '-0.2831881714E+05', '-0.2836665392E+05', '0.4783678177E+02']
        return np.array([float(strs_lst[aim_index].split("=")[1].split()[0].strip())])

    def get_atom_energies(self) -> np.ndarray | None:
        """Return the energies of individual atoms in material system.

        When turning on `ENERGY DEPOSITION`, PWmat will output energy per atom.

        Returns:
            np.ndarray | None: The energies of individual atoms within the material system.
        """
        energies = []
        aim_content = "Atomic-Energy, ".upper()
        aim_idxs = ListLocator.locate_all_lines(strs_lst=self.strs_lst, content=aim_content)
        if len(aim_idxs) == 0:
            return None
        aim_idx = aim_idxs[0]
        for tmp_str in self.strs_lst[aim_idx + 1 : aim_idx + self.num_atoms + 1]:
            """
            Atomic-Energy, Etot(eV),E_nonloc(eV),Q_atom:dE(eV)=  -0.1281163115E+06
            14   0.6022241483E+03    0.2413350871E+02    0.3710442365E+01
            """
            energies.append(float(tmp_str.split()[1]))
        return np.array(energies)

    def get_atom_forces(self) -> np.ndarray:
        """Return the force on atoms in material system.

        Returns:
            np.ndarray: Forces acting on individual atoms of shape=(num_atoms*3,)
        """
        aim_content = "Force".upper()
        aim_idx = ListLocator.locate_all_lines(strs_lst=self.strs_lst, content=aim_content, exclusion="average")[0]

        # ['14', '0.089910342901203', '0.077164252174742', '0.254144099204679']
        forces = [
            [float(val) for val in line.split()[1:4]]
            for line in self.strs_lst[aim_idx + 1 : aim_idx + self.num_atoms + 1]
        ]
        return -np.array(forces).reshape(-1)

    def get_virial(self) -> np.ndarray | None:
        """Return the virial tensor of material system.

        Returns:
            np.ndarray | None: Virial tensor of shape=(9,)
        """
        virial_tensor: list[float] = []
        aim_content = "LATTICE"
        aim_idx = ListLocator.locate_all_lines(strs_lst=self.strs_lst, content=aim_content)[0]

        for tmp_idx in [aim_idx + 1, aim_idx + 2, aim_idx + 3]:
            # tmp_strs_lst =
            # ['0.8759519000E+01', '0.0000000000E+00', '0.0000000000E+00',
            # 'stress', '(eV):', '0.115558E+02', '0.488108E+01', '0.238778E+01']
            tmp_strs_lst = self.strs_lst[tmp_idx].split()
            tmp_aim_row_lst = ListLocator.locate_all_lines(strs_lst=tmp_strs_lst, content="STRESS")
            if len(tmp_aim_row_lst) == 0:
                return None

        for tmp_idx in [aim_idx + 1, aim_idx + 2, aim_idx + 3]:
            # tmp_str_lst = ['0.120972E+02', '0.483925E+01', '0.242063E+01']
            tmp_str_lst = self.strs_lst[tmp_idx].split()[-3:]
            virial_tensor += (
                float(tmp_str_lst[0]),
                float(tmp_str_lst[1]),
                float(tmp_str_lst[2]),
            )

        return np.array(virial_tensor)


class AtomConfig(MSONable):
    """Object for representing the data in a atom.config or final.config file."""

    def __init__(self, structure: IStructure | Structure, sort_structure: bool = False):
        """Initialization function.

        Args:
            structure (Structure): Structure object
            sort_structure (bool, optional): Whether to sort the structure. Useful if species
                are not grouped properly together. Defaults to False.
        """
        self.structure: Structure = Structure.from_sites(structure)
        if sort_structure:
            self.structure = self.structure.get_sorted_structure()
        elements_counter = dict(sorted(Counter(self.structure.species).items()))
        true_names = [f"{tmp_key}{tmp_value}" for (tmp_key, tmp_value) in elements_counter.items()]
        self.true_names = "".join(true_names)

    def __repr__(self):
        return self.get_str()

    def __str__(self):
        return self.get_str()

    @classmethod
    def from_str(cls, data: str, mag: bool = False) -> Self:
        """Reads a atom.config from a string.

        Args:
            data (str): string containing atom.config data
            mag (bool, optional): Whether to read magnetic moment information.

        Returns:
            AtomConfig object
        """
        ac_extractor = ACstrExtractor(atom_config_str=data)
        properties: dict[str, float] = {}
        structure = Structure(
            lattice=ac_extractor.get_lattice(),
            species=ac_extractor.get_types(),  # type:ignore[arg-type]
            coords=ac_extractor.get_coords().reshape(-1, 3),
            coords_are_cartesian=False,
            properties=properties,
        )
        if mag:
            magmoms = ac_extractor.get_magmoms()
            for idx, tmp_site in enumerate(structure):
                tmp_site.properties |= {"magmom": magmoms[idx]}

        return cls(structure)

    @classmethod
    def from_file(cls, filename: PathLike, mag: bool = False) -> Self:
        """Get a AtomConfig from a file.

        Args:
            filename (PathLike): File name containing AtomConfig data
            mag (bool, optional): Whether to read magnetic moments. Defaults to True.

        Returns:
            AtomConfig object.
        """
        with zopen(filename, mode="rt", encoding="utf-8") as file:
            return cls.from_str(data=file.read(), mag=mag)  # type:ignore[arg-type]

    @classmethod
    def from_dict(cls, dct: dict) -> Self:
        """Get a AtomConfig object from a dictionary.

        Args:
            dct: dict containing atom.config data

        Returns:
            AtomConfig object.
        """
        return cls(Structure.from_dict(dct["structure"]))

    def get_str(self) -> str:
        """Return a string describing the structure in atom.config format.

        Returns:
            str: String representation of atom.config
        """
        # This corrects for VASP really annoying bug of crashing on lattices
        # which have triple product < 0. We will just invert the lattice vectors.
        lattice = self.structure.lattice
        if np.linalg.det(lattice.matrix) < 0:
            lattice = Lattice(-lattice.matrix)

        lines: list[str] = []
        lines += (f"\t{self.structure.num_sites} atoms\n", "Lattice vector\n")
        for idx in range(3):
            lines.append(f"{lattice.matrix[idx][0]:>15f}{lattice.matrix[idx][1]:>15f}{lattice.matrix[idx][2]:>15f}\n")
        lines.append("Position, move_x, move_y, move_z\n")
        for site_idx in range(len(self.structure)):
            lines += (
                f"{int(self.structure.species[site_idx].Z):>4d}",
                f"{self.structure.frac_coords[site_idx][0]:>15f}",
                f"{self.structure.frac_coords[site_idx][1]:>15f}",
                f"{self.structure.frac_coords[site_idx][2]:>15f}",
                "   1   1   1\n",
            )
        if "magmom" in self.structure.sites[0].properties:
            lines.append("MAGNETIC\n")
            for _, tmp_site in enumerate(self.structure.sites):
                lines.append(f"{int(tmp_site.specie.Z):>4d}{tmp_site.properties['magmom']:>15f}\n")
        return "".join(lines)

    def write_file(self, filename: PathLike, **kwargs):
        """Write AtomConfig to a file."""
        with zopen(filename, mode="wt", encoding="utf-8") as file:
            file.write(self.get_str(**kwargs))  # type:ignore[arg-type]

    def as_dict(self):
        """
        Returns:
            dict.
        """
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "structure": self.structure.as_dict(),
            "true_names": self.true_names,
        }


class GenKpt(MSONable):
    """Read and write gen.kpt. This file just generates line-mode kpoints."""

    def __init__(
        self,
        reciprocal_lattice: np.ndarray,
        kpoints: dict[str, np.ndarray],
        path: list[list[str]],
        density: float = 0.01,
    ) -> None:
        """Initialization function.

        Args:
            reciprocal_lattice (NDArray): Reciprocal lattice with factor of 2*pi.
            kpoints (dict[str, np.array]): Kpoints and their corresponding fractional coordinates.
            kpath (list[list[str]]): All kpaths, with each list representing one kpath.
            density (float): The density of kpoints mesh with factor of 2*pi.
        """
        self.reciprocal_lattice: np.ndarray = reciprocal_lattice
        self.kpath: dict = {}
        self.kpath |= {"kpoints": kpoints}
        self.kpath |= {"path": path}
        self.density = density

    @classmethod
    def from_structure(cls, structure: Structure | IStructure, dim: int, density: float = 0.01) -> Self:
        """Obtain a AtomConfig object from Structure object.

        Args:
            structure (Structure): A structure object.
            dim (int): The dimension of the material system (2 or 3).
            density (float): Kpoints mesh without factor with 2*pi. Program will
                automatically convert it with 2*pi.
        """
        kpath_set = KPathSeek(structure)
        if dim == 2:
            kpts_2d: dict[str, np.ndarray] = {}
            for tmp_name, tmp_kpt in kpath_set.kpath["kpoints"].items():
                if (tmp_kpt[2]) == 0:
                    kpts_2d |= {tmp_name: tmp_kpt}

            path_2d: list[list[str]] = []
            for tmp_path in kpath_set.kpath["path"]:
                tmp_path_2d: list[str] = [tmp_hsp for tmp_hsp in tmp_path if tmp_hsp in kpts_2d]

                if len(tmp_path_2d) > 1:
                    path_2d.append(tmp_path_2d)
            kpts: dict[str, np.ndarray] = kpts_2d
            path: list[list[str]] = path_2d
        else:
            kpts = kpath_set.kpath["kpoints"]
            path = kpath_set.kpath["path"]
        rec_lattice: np.ndarray = structure.lattice.reciprocal_lattice.matrix  # with 2*pi
        return cls(rec_lattice, kpts, path, density * 2 * np.pi)

    def get_str(self):
        """Get a string to be written as a gen.kpt file."""

        def calc_distance(hsp1: str, hsp2: str) -> float:
            """Calculate the distance between two high symmetry points.

            Args:
                hsp1 (str): The name of the first high symmetry point.
                hsp2 (str): The name of the second high symmetry point.

            Returns:
                float: The distance between two high symmetry points. With factor of 2*pi.
            """
            hsp1_coord: np.ndarray = np.dot(
                np.array(self.kpath["kpoints"][hsp1]).reshape(1, 3),
                self.reciprocal_lattice,
            )
            hsp2_coord: np.ndarray = np.dot(
                np.array(self.kpath["kpoints"][hsp2]).reshape(1, 3),
                self.reciprocal_lattice,
            )
            return float(np.linalg.norm(hsp2_coord - hsp1_coord))

        discontinue_pairs: list[list[int]] = [
            [self.kpath["path"][idx][-1], self.kpath["path"][idx + 1][0]] for idx in range(len(self.kpath["path"]) - 1)
        ]

        flatten_paths: list[str] = [tmp_hsp for tmp_path in self.kpath["path"] for tmp_hsp in tmp_path]
        gen_kpt_str: str = f"Generated by pymatgen. density={self.density / (2 * np.pi)}, "
        gen_kpt_str += "fractional coordinates in reciprocal lattice.\n"
        for idx in range(len(flatten_paths) - 1):
            if [flatten_paths[idx], flatten_paths[idx + 1]] not in discontinue_pairs:
                gen_kpt_str += f"{np.ceil(calc_distance(flatten_paths[idx], flatten_paths[idx + 1]) / self.density)}\n"
                gen_kpt_str += f"    {self.kpath['kpoints'][flatten_paths[idx]][0]:>12.6f}\t"
                gen_kpt_str += f"{self.kpath['kpoints'][flatten_paths[idx]][1]:>12.6f}\t"
                gen_kpt_str += f"{self.kpath['kpoints'][flatten_paths[idx]][2]:>12.6f}\t"
                gen_kpt_str += f"{flatten_paths[idx]}\n"
                gen_kpt_str += f"    {self.kpath['kpoints'][flatten_paths[idx + 1]][0]:>12.6f}\t"
                gen_kpt_str += f"{self.kpath['kpoints'][flatten_paths[idx + 1]][1]:>12.6f}\t"
                gen_kpt_str += f"{self.kpath['kpoints'][flatten_paths[idx + 1]][2]:>12.6f}\t"
                gen_kpt_str += f"{flatten_paths[idx + 1]}\n"
        return gen_kpt_str

    def write_file(self, filename: PathLike):
        """Write gen.kpt to a file.

        Args:
            filename (PathLike): The absolute path of file to be written.
        """
        with zopen(filename, mode="wt", encoding="utf-8") as file:
            file.write(self.get_str())


class HighSymmetryPoint(MSONable):
    """Read and write HIGH_SYMMETRY_POINTS file which generate line-mode kpoints."""

    def __init__(
        self,
        reciprocal_lattice: np.ndarray,
        kpts: dict[str, list],
        path: list[list[str]],
        density: float,
    ):
        """Initialization function.

        Args:
            reciprocal_lattice (np.array): Reciprocal lattice.
            kpts (dict[str, list[float]]): Kpoints and their corresponding fractional coordinates.
            path (list[list[str]]): All k-paths, with each list representing one k-path.
            density (float): Density of kpoints mesh with factor of 2*pi.
        """
        self.reciprocal_lattice: np.ndarray = reciprocal_lattice
        self.kpath: dict = {}
        self.kpath |= {"kpoints": kpts}
        self.kpath |= {"path": path}
        self.density = density

    @classmethod
    def from_structure(cls, structure: Structure | IStructure, dim: int, density: float = 0.01) -> Self:
        """Obtain HighSymmetry object from Structure object.

        Args:
            structure (Structure): A structure object.
            dim (int): Dimension of the material system (2 or 3).
            density (float, optional): Density of kpoints mesh without factor of 2*pi. Defaults to 0.01.
                The program will automatically convert it to with factor of 2*pi.
        """
        reciprocal_lattice: np.ndarray = structure.lattice.reciprocal_lattice.matrix
        gen_kpt = GenKpt.from_structure(structure=structure, dim=dim, density=density)
        return cls(
            reciprocal_lattice,
            gen_kpt.kpath["kpoints"],
            gen_kpt.kpath["path"],
            density * 2 * np.pi,
        )

    def get_str(self) -> str:
        """Get a string describing high symmetry points in HIGH_SYMMETRY_POINTS format."""

        def calc_distance(hsp1: str, hsp2: str) -> float:
            """Calculate the distance of two high symmetry points.

            Returns:
                float: The distance between two high symmetry points with factor of 2*pi.
            """
            hsp1_coord: np.ndarray = np.dot(
                np.array(self.kpath["kpoints"][hsp1]).reshape(1, 3),
                self.reciprocal_lattice,
            )
            hsp2_coord: np.ndarray = np.dot(
                np.array(self.kpath["kpoints"][hsp2]).reshape(1, 3),
                self.reciprocal_lattice,
            )
            return float(np.linalg.norm(hsp2_coord - hsp1_coord))

        def get_hsp_row_str(label: str, index: int, coordinate: float) -> str:
            """
            Return string containing name, index, coordinate of the certain high symmetry point
            in HIGH_SYMMETRY_POINTS format.

            Args:
                label (str): Name of the high symmetry point.
                index (int): Index of the high symmetry point.
                coordinate (float): Coordinate in bandstructure of the high symmetry point.

            Returns:
                str: String containing name, index, coordinate of the certain high symmetry point
                    in HIGH_SYMMETRY_POINTS format.
            """
            if label == "GAMMA":
                return f"G            {index:>4d}         {coordinate:>.6f}\n"
            return f"{label}            {index:>4d}         {coordinate:>.6f}\n"

        discontinue_pairs: list[list[str]] = [
            [self.kpath["path"][ii][-1], self.kpath["path"][ii + 1][0]] for ii in range(len(self.kpath["path"]) - 1)
        ]

        flatten_paths: list[str] = [tmp_hsp for tmp_path in self.kpath["path"] for tmp_hsp in tmp_path]
        # flatten_paths = [hsp.replace("GAMMA", "G") for hsp in flatten_paths]

        index: int = 1
        coordinate: float = 0.0
        hsp_str: str = "Label       Index       Coordinate\n"
        hsp_str += get_hsp_row_str(flatten_paths[0], index, coordinate)
        for ii in range(1, len(flatten_paths)):
            if [flatten_paths[ii - 1], flatten_paths[ii]] not in discontinue_pairs:
                coordinate += calc_distance(flatten_paths[ii - 1], flatten_paths[ii]) / (2 * np.pi)
                index += int(np.ceil(calc_distance(flatten_paths[ii - 1], flatten_paths[ii]) / self.density + 1))
            else:
                coordinate += 0.0
                index += 1
            hsp_str += get_hsp_row_str(flatten_paths[ii], index, coordinate)
        return hsp_str

    def write_file(self, filename: PathLike):
        """Write HighSymmetryPoint to a file."""
        with zopen(filename, mode="wt", encoding="utf-8") as file:
            file.write(self.get_str())  # type:ignore[arg-type]



"""Module for JDFTx IO module output utils.

Module for JDFTx IO module output utils. Functions kept in this module are here if they are
used by multiple submodules, or if they are anticipated to be used by multiple
submodules in the future.
"""

from __future__ import annotations

from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeAlias

import numpy as np

from pymatgen.electronic_structure.core import Orbital

if TYPE_CHECKING:
    from collections.abc import Callable

    from numpy.typing import NDArray


def check_file_exists(func: Callable) -> Any:
    """Check if file exists.

    Check if file exists (and continue normally) or raise an exception if
    it does not.
    """

    @wraps(func)
    def wrapper(filename: str) -> Any:
        filepath = Path(filename)
        if not filepath.is_file():
            raise OSError(f"'{filename}' file doesn't exist!")
        return func(filename)

    return wrapper


@check_file_exists
def read_file(file_name: str) -> list[str]:
    """
    Read file into a list of str.

    Parameters
    ----------
    filename: Path or str
        name of file to read

    Returns
    -------
    text: list[str]
        list of strings from file
    """
    with open(file_name) as f:
        text = f.readlines()
    f.close()
    return text


def read_outfile_slices(file_name: str) -> list[list[str]]:
    """
    Read slice of out file into a list of str.

    Parameters
    ----------
    filename: Path or str
        name of file to read
    out_slice_idx: int
        index of slice to read from file

    Returns
    -------
    texts: list[list[str]]
        list of out file slices (individual calls of JDFTx)
    """
    _text = read_file(file_name)
    start_lines = get_start_lines(_text, add_end=True)
    texts = []
    for i in range(len(start_lines) - 1):
        text = _text[start_lines[i] : start_lines[i + 1]]
        texts.append(text)
    return texts


def _brkt_list_of_3_to_nparray(line: str) -> NDArray[np.float64]:
    """Return 3x1 numpy array.

    Convert a string of the form "[ x y z ]" to a 3x1 numpy array

    Parameters
    ----------
    line: str
        A string of the form "[ x y z ]"
    """
    return np.array([float(x) for x in line.split()[1:-1]])


def _brkt_list_of_3x3_to_nparray(lines: list[str], i_start: int = 0) -> NDArray[np.float64]:
    """Return 3x3 numpy array.

    Convert a list of strings of the form "[ x y z ]" to a 3x3 numpy array

    Parameters
    ----------
    lines: list[str]
        A list of strings of the form "[ x y z ]"
    i_start: int
        The index of the first line in lines

    Returns
    -------
    out: np.ndarray
        A 3x3 numpy array
    """
    out = np.zeros([3, 3])
    for i in range(3):
        out[i, :] += _brkt_list_of_3_to_nparray(lines[i + i_start])
    return out


# Named "t1" in unmet anticipation of multiple ways that a float would be needed
# to be read following the variable string with a colon.
def get_colon_val(linetext: str, lkey: str) -> float | np.float64 | None:
    """Return float val from '...lkey: val...' in linetext.

    Read a float from an elec minimization line assuming value appears as
    "... lkey value ...".

    Parameters
    ----------
    linetext: str
        A line of text from a JDFTx out file
    lkey: str
        A string that appears before the float value in linetext. Must include
        the colon.
    """
    colon_var = None
    if lkey in linetext:
        val = linetext.split(lkey)[1].strip().split(" ")[0]
        colon_var = np.nan if val == "nan" else float(linetext.split(lkey)[1].strip().split(" ")[0])
    return colon_var


# Temporary alias until the outside modules are merged with the renaming
get_colon_var_t1 = get_colon_val


# This function matches the format of the generic "is_<x>_start_line" functions specific to
# the JOutStructure object initialization, but is not moved to the joutstructure module
# as it is also used in methods for other JDFTx IO modules (e.g. JOutStructures) so it
# is kept here to avoid circular imports.
def is_lowdin_start_line(line_text: str) -> bool:
    """Return True if the line_text is the start of Lowdin log message.

    Return True if the line_text is the start of a Lowdin population analysis
    in a JDFTx out file.

    Parameters
    ----------
    line_text: str
        A line of text from a JDFTx out file

    Returns
    -------
    is_line: bool
        True if the line_text is the start of a Lowdin population analysis in a
        JDFTx out file
    """
    return "#--- Lowdin population analysis ---" in line_text


# TODO: Figure out if this ever actually gets called, (I think I added it to make JOutStructure user friendly
# but JOutStructure is not intended to be a user-initialized class)
def correct_geom_opt_type(opt_type: str | None) -> str | None:
    """Return recognizable opt_type string.

    Correct the opt_type string to match the JDFTx convention.

    Parameters
    ----------
    opt_type:
        The type of optimization step

    Returns
    -------
    opt_type: str | None
        The corrected type of optimization step
    """
    if opt_type is not None:
        if "lattice" in opt_type.lower():
            opt_type = "LatticeMinimize"
        elif "ionic" in opt_type.lower():
            opt_type = "IonicDynamics" if "dyn" in opt_type.lower() else "IonicMinimize"
        else:
            opt_type = None
    return opt_type


def get_start_lines(
    text: list[str],
    start_key: str = "*************** JDFTx",
    add_end: bool = False,
) -> list[int]:
    """Get start line numbers for JDFTx calculations.

    Get the line numbers corresponding to the beginning of separate JDFTx calculations
    (in case of multiple calculations appending the same out file).

    Args:
        text: output of read_file for out file
    """
    start_lines = []
    i = None
    for i, line in enumerate(text):
        if start_key in line:
            start_lines.append(i)
    if add_end and i is not None:
        start_lines.append(i)
    if i is None:
        raise ValueError("Outfile parser fed an empty file.")
    if not start_lines:
        raise ValueError("No JDFTx calculations found in file.")
    return start_lines


def find_key_first(key_input: str, tempfile: list[str]) -> int | None:
    """Find the first instance of a key in the output file.

    Args:
        key_input (str): Key string to match.
        tempfile (list[str]): Output from readlines() function in read_file method.

    Returns:
        int | None: The index of the first occurrence of the key in the tempfile list, or None if the key is not found.
    """
    key_input = str(key_input)
    line = None
    for i in range(len(tempfile)):
        if key_input in tempfile[i]:
            line = i
            break
    return line


def find_key(key_input: str, tempfile: list[str]) -> int | None:
    """
    Find the last instance of a key in the output file.

    Args:
        key_input (str): Key string to match.
        tempfile (list[str]): Output from readlines() function in read_file method.

    Returns:
        int | None: The index of the last occurrence of the key in the tempfile list, or None if the key is not found.
    """
    key_input = str(key_input)
    line = None
    lines = find_all_key(key_input, tempfile)
    if len(lines):
        line = lines[-1]
    return line


def find_first_range_key(
    key_input: str,
    tempfile: list[str],
    startline: int = 0,
    endline: int = -1,
    skip_pound: bool = False,
) -> list[int]:
    """
    Find all lines that exactly begin with key_input in a range of lines.

    Args:
        key_input (str): Key string to match.
        tempfile (list[str]): Output from readlines() function in read_file method.
        startline (int): Line to start searching from.
        endline (int): Line to stop searching at.
        skip_pound (bool): Whether to skip lines that begin with a pound sign.

    Returns:
        list[int]: List of line numbers where key_input occurs.
    """
    key_input = str(key_input)
    startlen = len(key_input)
    line_list = []

    if endline == -1:
        endline = len(tempfile)
    for i in range(startline, endline):
        line = tempfile[i]
        if skip_pound:
            for _ in range(10):  # repeat to make sure no really weird formatting
                line = line.lstrip()
                line = line.lstrip("#")
        line = line[0:startlen]
        if line == key_input:
            line_list.append(i)
    return line_list


def key_exists(key_input: str, tempfile: list[str]) -> bool:
    """
    Check if key_input exists in tempfile.

    Search through tempfile for key_input. Return True if found, False otherwise.

    Args:
        key_input (str): Key string to match.
        tempfile (list[str]): Output from readlines() function in read_file method.

    Returns:
        bool: True if key_input exists in tempfile, False otherwise.
    """
    line = find_key(key_input, tempfile)
    return line is not None


def find_all_key(key_input: str, tempfile: list[str], startline: int = 0) -> list[int]:
    """Find all lines containing key_input.

    Search through tempfile for all lines containing key_input. Returns a list of line numbers.

    Args:
        key_input (str): Key string to match.
        tempfile (list[str]): Output from readlines() function in read_file method.
        startline (int): Line to start searching from.

    Returns:
        list[int]: List of line numbers where key_input occurs.
    """
    return [i for i in range(startline, len(tempfile)) if key_input in tempfile[i]]


def _init_dict_from_colon_dump_lines(lines: list[str]):
    varsdict = {}
    for line in lines:
        if ":" in line:
            lsplit = line.split(":")
            key = lsplit[0].strip()
            val = lsplit[1].split()[0].strip()
            varsdict[key] = val
    return varsdict


def _parse_bandfile_complex(bandfile_filepath: str | Path) -> NDArray[np.complex64]:
    Dtype: TypeAlias = np.complex64
    token_parser = _complex_token_parser
    return _parse_bandfile_reader(bandfile_filepath, Dtype, token_parser)


def _parse_bandfile_normalized(bandfile_filepath: str | Path) -> NDArray[np.float32]:
    Dtype: TypeAlias = np.float32
    token_parser = _normalized_token_parser
    return _parse_bandfile_reader(bandfile_filepath, Dtype, token_parser)


def _get__from_bandfile_filepath(bandfile_filepath: Path | str, tok_idx: int) -> int:
    """
    Get arbitrary integer from header of bandprojections file.

    Args:
        bandfile_filepath (Path | str): Path to bandprojections file.
        tok_idx (int): Index of token to return.

    Returns:
        int: Integer from header of bandprojections file.
    """
    ret_data = None
    bandfile = read_file(bandfile_filepath)
    for iLine, line in enumerate(bandfile):
        tokens = line.split()
        if iLine == 0:
            ret_data = int(tokens[tok_idx])
            break
    if ret_data is None:
        raise ValueError("Provided an empty file")
    return ret_data


def _get_nstates_from_bandfile_filepath(bandfile_filepath: Path | str) -> int:
    """Get number of states from bandprojections file.

    Get the number of states from the bandprojections file.

    Args:
        bandfile_filepath (Path | str): Path to bandprojections file.

    Returns:
        int: Number of states.
    """
    return _get__from_bandfile_filepath(bandfile_filepath, 0)


def _get_nbands_from_bandfile_filepath(bandfile_filepath: Path | str) -> int:
    """Get number of bands from bandprojections file.

    Get the number of bands from the bandprojections file. The output here should match up with `nbands` from the
    JDFTXOutfile object.

    Args:
        bandfile_filepath (Path | str): Path to bandprojections file.

    Returns:
        int: Number of bands.
    """
    return _get__from_bandfile_filepath(bandfile_filepath, 2)


def _get_nproj_from_bandfile_filepath(bandfile_filepath: Path | str) -> int:
    """Get number of projections from bandprojections file.

    Get the number of projections from the bandprojections file.

    Args:
        bandfile_filepath (Path | str): Path to bandprojections file.

    Returns:
        int: Number of projections.
    """
    return _get__from_bandfile_filepath(bandfile_filepath, 4)


def _get_nspecies_from_bandfile_filepath(bandfile_filepath: Path | str) -> int:
    """Get number of species (ion types) from bandprojections file.

    Get the number of species (ion types) from the bandprojections file.

    Args:
        bandfile_filepath (Path | str): Path to bandprojections file.

    Returns:
        int: Number of species.
    """
    return _get__from_bandfile_filepath(bandfile_filepath, 6)


def _parse_bandfile_reader(bandfile_filepath: str | Path, arr_dtype: TypeAlias, token_parser: Callable) -> NDArray:
    nstates = _get_nstates_from_bandfile_filepath(bandfile_filepath)
    nbands = _get_nbands_from_bandfile_filepath(bandfile_filepath)
    nproj = _get_nproj_from_bandfile_filepath(bandfile_filepath)
    nspecies = _get_nspecies_from_bandfile_filepath(bandfile_filepath)
    # Header of length 3, and then each states occupies 1 (header) + nbands lineas
    bandfile = read_file(bandfile_filepath)
    expected_length = 2 + nspecies + (nstates * (1 + nbands))
    if not expected_length == len(bandfile):
        raise RuntimeError("Bandprojections file does not match expected length - ensure no edits have been made.")
    proj_tju: NDArray[arr_dtype] = np.zeros((nstates, nbands, nproj), dtype=arr_dtype)
    for line, text in enumerate(bandfile):
        tokens = text.split()
        if line >= nspecies + 2:
            istate = (line - (nspecies + 2)) // (nbands + 1)
            iband = (line - (nspecies + 2)) - istate * (nbands + 1) - 1
            if iband >= 0 and istate < nstates:
                proj_tju[istate, iband] = np.array(token_parser(tokens))
    return proj_tju


def _complex_token_parser(tokens: list[str]) -> NDArray[np.complex64]:
    out = np.zeros(int(len(tokens) / 2), dtype=np.complex64)
    ftokens = np.array(tokens, dtype=np.float32)
    out += 1j * ftokens[1::2]
    out += ftokens[::2]
    return out


def _normalized_token_parser(tokens: list[str]) -> NDArray[np.float32]:
    normalized_tokens: NDArray[np.float32] = np.array(tokens, dtype=np.float32)
    return normalized_tokens


def get_proj_tju_from_file(bandfile_filepath: Path | str) -> NDArray[np.float32 | np.complex64]:
    """Return projections from file in tju shape.

    Return projections from file in (state, band, proj) shape. Collected in this shape before sabcju shape due to ready
    availability of this shape in the file.

    Args:
        bandfile_filepath (Path | str): Path to bandprojections file.

    Returns:
        np.ndarray: Projections array in shape (state, band, proj).
    """
    is_complex = _is_complex_bandfile_filepath(bandfile_filepath)
    return _parse_bandfile_complex(bandfile_filepath) if is_complex else _parse_bandfile_normalized(bandfile_filepath)


def _parse_kptsfrom_bandprojections_file(bandfile_filepath: str | Path) -> tuple[list[float], list[NDArray]]:
    """Parse kpts from bandprojections file.

    Parse kpts from bandprojections file.

    Args:
        bandfile_filepath (Path | str): Path to bandprojections file.

    Returns:
        tuple[list[float], list[np.ndarray[float]]]: Tuple of k-point weights and k-points
    """
    wk_list: list[float] = []
    k_points_list: list[NDArray] = []
    kpt_lines = []
    with open(bandfile_filepath) as f:
        for line in f:
            if line.startswith("#") and ";" in line:
                _line = line.split(";")[0].lstrip("#")
                kpt_lines.append(_line)
    for line in kpt_lines:
        k_points = line.split("[")[1].split("]")[0].strip().split()
        _k_points_floats: list[float] = [float(v) for v in k_points]
        k_points_list.append(np.array(_k_points_floats))
        wk = float(line.split("]")[1].strip().split()[0])
        wk_list.append(wk)
    return wk_list, k_points_list


def _is_complex_bandfile_filepath(bandfile_filepath: str | Path) -> bool:
    """Determine if bandprojections file is complex.

    Determine if the bandprojections file is complex. Needed before attempting pCOHP analysis.

    Args:
        bandfile_filepath (Path | str): Path to bandprojections file.

    Returns:
        bool: True if the bandprojections file is complex, False otherwise.
    """
    hash_lines = 0
    val = True
    with open(bandfile_filepath) as f:
        for _i, line in enumerate(f):
            if "#" in line:
                hash_lines += 1
                if hash_lines == 2:
                    val = "|projection|^2" not in line
                    break
    f.close()
    return val


# TODO: This is very likely redundant to something in pymatgen - replace with that if possible.
orb_ref_list = [
    ["s"],
    ["py", "pz", "px"],
    ["dxy", "dyz", "dz2", "dxz", "dx2-y2"],
    ["fy(3x2-y2)", "fxyz", "fyz2", "fz3", "fxz2", "fz(x2-y2)", "fx(x2-3y2)"],
]
orb_ref_to_o_dict = {
    "s": int(Orbital.s),
    "py": int(Orbital.py),
    "pz": int(Orbital.pz),
    "px": int(Orbital.px),
    "dxy": int(Orbital.dxy),
    "dyz": int(Orbital.dyz),
    "dz2": int(Orbital.dz2),
    "dxz": int(Orbital.dxz),
    "dx2-y2": int(Orbital.dx2),
    # Keep the f-orbitals arbitrary-ish until they get designated names in pymatgen.
    orb_ref_list[-1][0]: int(Orbital.f_3),
    orb_ref_list[-1][1]: int(Orbital.f_2),
    orb_ref_list[-1][2]: int(Orbital.f_1),
    orb_ref_list[-1][3]: int(Orbital.f0),
    orb_ref_list[-1][4]: int(Orbital.f1),
    orb_ref_list[-1][5]: int(Orbital.f2),
}


def _get_atom_orb_labels_map_dict(bandfile_filepath: Path) -> dict[str, list[str]]:
    """
    Return a dictionary mapping each atom symbol to pymatgen-compatible orbital projection string representations.

    Identical to _get_atom_orb_labels_ref_dict, but doesn't include the numbers in the labels.



    Args:
        bandfile_filepath (str | Path): The path to the bandfile.

    Returns:
        dict[str, list[str]]: A dictionary mapping each atom symbol to all atomic orbital projection string
        representations.
    """
    bandfile = read_file(bandfile_filepath)
    labels_dict: dict[str, list[str]] = {}

    for i, line in enumerate(bandfile):
        if i > 1:
            if "#" in line:
                break
            lsplit = line.strip().split()
            sym = lsplit[0]
            labels_dict[sym] = []
            lmax = int(lsplit[3])
            # Would prefer to use "l" rather than "L" here (as uppercase "L" means something else entirely) but
            # pr*-c*mm*t thinks "l" is an ambiguous variable name.
            for L in range(lmax + 1):
                mls = orb_ref_list[L]
                nshells = int(lsplit[4 + L])
                for _n in range(nshells):
                    if nshells > 1:
                        for ml in mls:
                            labels_dict[sym].append(f"{ml}")
                    else:
                        labels_dict[sym] += mls
    return labels_dict


def _get_atom_orb_labels_ref_dict(bandfile_filepath: Path) -> dict[str, list[str]]:
    """
    Return a dictionary mapping each atom symbol to all atomic orbital projection string representations.

    Return a dictionary mapping each atom symbol to all atomic orbital projection string representations.
    For example:
    {
        "H": ["s"],
        "O": ["s", "px", "py", "pz", "dxy", "dxz", "dyz", "dx2y2", "dz2"],
        "Pt": ["0s", "1s", "0px", "0py", "0pz", "1px", "1py", "1pz", "dxy", "dxz", "dyz", "dx2y2", "dz2",
               "fx3-3xy2", "fyx2-yz2", "fxz2", "fz3", "fyz2", "fxyz", "f3yx2-y3"]
    }
    where the numbers are needed when using pseudopotentials with multiple valence shells of the same angular momentum
    and are NOT REPRESENTATIVE OF THE TRUE PRINCIPAL QUANTUM NUMBER.

    Args:
        bandfile_filepath (str | Path): The path to the bandfile.

    Returns:
        dict[str, list[str]]: A dictionary mapping each atom symbol to all atomic orbital projection string
        representations.
    """
    bandfile = read_file(bandfile_filepath)
    labels_dict: dict[str, list[str]] = {}

    for i, line in enumerate(bandfile):
        if i > 1:
            if "#" in line:
                break
            lsplit = line.strip().split()
            sym = lsplit[0]
            labels_dict[sym] = []
            lmax = int(lsplit[3])
            # Would prefer to use "l" rather than "L" here (as uppercase "L" means something else entirely) but
            # pr*-c*mm*t thinks "l" is an ambiguous variable name.
            for L in range(lmax + 1):
                mls = orb_ref_list[L]
                nshells = int(lsplit[4 + L])
                for n in range(nshells):
                    if nshells > 1:
                        for ml in mls:
                            labels_dict[sym].append(f"{n}{ml}")
                    else:
                        labels_dict[sym] += mls
    return labels_dict


def _get_atom_count_list(bandfile_filepath: Path) -> list[tuple[str, int]]:
    """
    Return a list of tuples of atom symbols and counts.

    Return a list of tuples of atom symbols and counts. This is superior to a dictionary as it maintains the order of
    the atoms in the bandfile.

    Args:
        bandfile_filepath (str | Path): The path to the bandfile.

    Returns:
        list[tuple[str, int]]: A list of tuples of atom symbols and counts.
    """
    bandfile = read_file(bandfile_filepath)
    atom_count_list = []

    for i, line in enumerate(bandfile):
        if i > 1:
            if "#" in line:
                break
            lsplit = line.strip().split()
            sym = lsplit[0].strip()
            count = int(lsplit[1].strip())
            atom_count_list.append((sym, count))
    return atom_count_list


def _get_orb_label_list_expected_len(labels_dict: dict[str, list[str]], atom_count_list: list[tuple[str, int]]) -> int:
    """
    Return the expected length of the atomic orbital projection string representation list.

    Return the expected length of the atomic orbital projection string representation list.

    Args:
        labels_dict (dict[str, list[str]]): A dictionary mapping each atom symbol to all atomic orbital projection
        string representations.
        atom_count_list (list[tuple[str, int]]): A list of tuples of atom symbols and counts.

    Returns:
        int: The expected length of the atomic orbital projection string representation list.
    """
    expected_len = 0
    for ion_tuple in atom_count_list:
        ion = ion_tuple[0]
        count = ion_tuple[1]
        orbs = labels_dict[ion]
        expected_len += count * len(orbs)
    return expected_len


def _get_orb_label(ion: str, idx: int, orb: str) -> str:
    """
    Return the string representation for an orbital projection.

    Return the string representation for an orbital projection.

    Args:
        ion (str): The symbol of the atom.
        idx (int): The index of the atom.
        orb (str): The atomic orbital projection string representation.

    Returns:
        str: The atomic orbital projection string representation for the atom.
    """
    return f"{ion}#{idx + 1}({orb})"


def _get_u_to_oa_map(bandfile_filepath: Path) -> list[tuple[int, int]]:
    """
    Return a list, where the u'th element is a tuple of the atomic orbital index and the ion index.

    Args:
        bandfile_filepath (str | Path): The path to the bandfile.

    Returns:
        list[tuple[int, int]]: A list, where the u'th element is a tuple of the atomic orbital index and the ion index.
    """
    map_labels_dict = _get_atom_orb_labels_map_dict(bandfile_filepath)
    atom_count_list = _get_atom_count_list(bandfile_filepath)
    u_to_oa_map: list = []
    a = 0
    for ion, ion_count in atom_count_list:
        for _i in range(ion_count):
            u_to_oa_map.extend((orb_ref_to_o_dict[orb], a) for orb in map_labels_dict[ion])
            a += 1
    return u_to_oa_map


def _get_orb_label_list(bandfile_filepath: Path) -> tuple[str, ...]:
    """
    Return a tuple of all atomic orbital projection string representations.

    Return a tuple of all atomic orbital projection string representations.

    Args:
        bandfile_filepath (str | Path): The path to the bandfile.

    Returns:
        tuple[str]: A list of all atomic orbital projection string representations.
    """
    labels_dict = _get_atom_orb_labels_ref_dict(bandfile_filepath)
    atom_count_list = _get_atom_count_list(bandfile_filepath)
    read_file(bandfile_filepath)
    labels_list: list[str] = []
    for ion_tuple in atom_count_list:
        ion = ion_tuple[0]
        orbs = labels_dict[ion]
        count = ion_tuple[1]
        labels_list.extend(_get_orb_label(ion, i, orb) for i in range(count) for orb in orbs)
    # This is most likely unnecessary, but it is a good check to have.
    if len(labels_list) != _get_orb_label_list_expected_len(labels_dict, atom_count_list):
        raise RuntimeError("Number of atomic orbital projections does not match expected length.")
    return tuple(labels_list)



"""Stub file for a guaranteed safe import of duecredit constructs: if duecredit
is not available.

Then use in your code as

    from .due import due, Doi, BibTeX, Text

See https://github.com/duecredit/duecredit/blob/master/README.md for examples.

Origin:     Originally a part of the duecredit
Copyright:  2015-2021 DueCredit developers
License:    BSD-2
"""

from __future__ import annotations

import logging

__version__ = "0.0.9"


class InactiveDueCreditCollector:
    """Just a stub at the Collector which would not do anything."""

    def _donothing(self, *args, **kwargs):
        """Perform no good and no bad."""

    def dcite(self, *args, **kwargs):
        """If I could cite I would."""

        def non_decorating_decorator(func):
            return func

        return non_decorating_decorator

    active = False
    activate = add = cite = dump = load = _donothing

    def __repr__(self):
        return f"{type(self).__name__}()"


def _donothing_func(*args, **kwargs):
    """Perform no good and no bad."""


try:
    from duecredit import BibTeX, Doi, Text, Url, due

    if "due" in locals() and not hasattr(due, "cite"):
        raise RuntimeError("Imported due lacks .cite. DueCredit is now disabled")
except Exception as exc:
    if not isinstance(exc, ImportError):
        logging.getLogger("duecredit").exception("Failed to import duecredit")
    # Initiate due stub
    due = InactiveDueCreditCollector()
    BibTeX = Doi = Url = Text = _donothing_func



"""
This module provides a pymatgen I/O interface to PACKMOL.

- PackmolSet provides a "run" method to run PACKMOL locally.
- PackmolBoxGen provides "get_input_set" for packing molecules into a box,
which returns a PackmolSet object.

For the run() method to work, you need to install the PACKMOL package.
See https://m3g.iqm.unicamp.br/packmol for download and setup instructions.
After installation, you may need to add the path of the PACKMOL
executable to the PATH environment variable.

Note that PACKMOL versions prior to 20.3.0 do not support paths with spaces.
"""

from __future__ import annotations

import logging
import subprocess
from pathlib import Path
from shutil import which
from typing import TYPE_CHECKING

import numpy as np
from monty.os import cd

from pymatgen.core import Molecule
from pymatgen.io.core import InputGenerator, InputSet

if TYPE_CHECKING:
    from pymatgen.util.typing import PathLike

__author__ = "Tingzheng Hou, Ryan Kingsbury, Orion Cohen"
__version__ = "1.0"
__maintainer__ = "Ryan Kingsbury"
__email__ = "RKingsbury@lbl.gov"
__date__ = "Nov 2021"

logger = logging.getLogger(__name__)

PACKMOL_CONSTRAINTS = (
    "fixed",
    "inside cube",
    "outside cube",
    "inside box",
    "outside box",
    "inside sphere",
    "outside sphere",
    "inside ellipsoid",
    "outside ellipsoid",
    "above plane",
    "below plane",
    "inside cylinder",
    "outside cylinder",
    "constrain_rotation",
    "over xygauss",
    "below xygauss",
)


class PackmolSet(InputSet):
    """InputSet for the PACKMOL software. This class defines several attributes related to."""

    def run(self, path: PathLike, timeout: float = 30) -> None:
        """Run PACKMOL and write out the packed structure.

        Args:
            path (PathLike): The path in which packmol input files are located.
            timeout (float): Timeout in seconds.

        Raises:
            ValueError: if packmol does not succeed in packing the box.
            TimeoutExpiredError: if packmol does not finish within the timeout.
        """
        if not which("packmol"):
            raise RuntimeError(
                "Running a PackmolSet requires the executable 'packmol' to be in the path. Please "
                "download packmol from https://github.com/leandromartinez98/packmol and follow the "
                "instructions in the README to compile. Don't forget to add the packmol binary to your path"
            )
        with cd(path):
            try:
                with open(self.inputfile, encoding="utf-8") as infile:
                    proc = subprocess.run(
                        ["packmol"],
                        stdin=infile,
                        check=True,
                        timeout=timeout,
                        capture_output=True,
                    )
                # This workaround is needed because packmol can fail to find
                # a solution but still return a zero exit code.
                # See https://github.com/m3g/packmol/issues/28
                if "ERROR" in proc.stdout.decode():
                    if "Could not open file." in proc.stdout.decode():
                        raise ValueError(
                            "Your packmol might be too old to handle paths with spaces."
                            "Please try again with a newer version or use paths without spaces."
                        )
                    msg = proc.stdout.decode().split("ERROR")[-1]
                    raise ValueError(f"Packmol failed with return code 0 and stdout: {msg}")

            except subprocess.CalledProcessError as exc:
                raise ValueError(f"Packmol failed with error code {exc.returncode} and stderr: {exc.stderr}") from exc
            else:
                with open(self.stdoutfile, mode="w", encoding="utf-8") as out:
                    out.write(proc.stdout.decode())

    @classmethod
    def from_directory(cls, directory: PathLike) -> None:
        """
        Construct an InputSet from a directory of one or more files.

        Args:
            directory (PathLike): Directory to read input files from.
        """
        raise NotImplementedError(f"from_directory has not been implemented in {cls.__name__}")


class PackmolBoxGen(InputGenerator):
    """
    Generator for a Packmol InputSet that packs one or more molecules into a rectangular
    simulation box.
    """

    def __init__(
        self,
        tolerance: float = 2.0,
        seed: int = 1,
        control_params: dict | None = None,
        inputfile: PathLike = "packmol.inp",
        outputfile: PathLike = "packmol_out.xyz",
        stdoutfile: PathLike = "packmol.stdout",
    ) -> None:
        """
        Instantiate a PackmolBoxGen class. The init method defines simulations parameters
        like filenames, random seed, tolerance, etc.

        Args:
            tolerance (float): Tolerance for packmol, in Å.
            seed (int): Random seed for packmol. Use 1 (default) for deterministic
                output, or -1 to generate a new random seed from the current time.
            control_params (dict): Additional parameters to be added to the input file.
            inputfile (PathLike): Path to the input file. Default to "packmol.inp".
            outputfile (PathLike): Path to the output file. Default to "packmol_out.xyz".
            stdoutfile (PathLike): Path to the file where stdout will be recorded. Default to "packmol.stdout".
        """
        self.inputfile = inputfile
        self.outputfile = outputfile
        self.stdoutfile = stdoutfile
        self.control_params = control_params or {}
        self.tolerance = tolerance
        self.seed = seed

    def get_input_set(
        self,
        molecules: list[dict],
        box: list[float] | None = None,
    ) -> PackmolSet:
        """Generate a Packmol InputSet for a set of molecules.

        Args:
            molecules (list[dict]): Information about molecules to pack
                into the box. Each dict requires three keys + 2 optional keys:
                    1. "name" - the structure name.
                    2. "number" - the number of that molecule to pack into the box.
                    3. "coords" - Coordinates in the form of either a Molecule
                        object or a path to a file.
                    4. "constraints" - (optional) the list of constraints for this
                        group of molecules.
                    5. "atoms_constraints" - (optional) the list of constraints for
                        specific atoms of this group of molecules. Each item in the
                        list is a dictionary with keys 'indices' and 'constraints'.
                        The 'indices' key contains the atom indices (0-based) of the
                        atoms to which the list of 'constraints' applies.
                Example:
                    {
                        "name": "water",
                        "number": 500,
                        "coords": "/path/to/input/file.xyz",
                    }
            box (list[float]): Box dimensions xlo, ylo, zlo, xhi, yhi, zhi, in Å. If set to None
                (default), pymatgen will estimate the required box size based on the volumes of
                the provided molecules.

        Returns:
            PackmolSet
        """
        mapping: dict = {}
        names = [d["name"] for d in molecules]
        if len(names) != len(set(names)):
            raise ValueError("Names of the molecules should all be different.")
        file_contents: list[str] = [
            "# Packmol input generated by pymatgen.\n",
            f"# {' + '.join(str(d['number']) + ' ' + d['name'] for d in molecules)}",
        ]

        for key, val in self.control_params.items():
            if isinstance(val, list):
                file_contents.append(f"{key} {' '.join(str(x) for x in val)}")
            else:
                file_contents.append(f"{key} {val}")

        file_contents += [
            f"seed {self.seed}",
            f"tolerance {self.tolerance}\n",
            "filetype xyz\n",
        ]

        if " " in str(self.outputfile):
            # NOTE - double quotes are deliberately used inside the f-string here, do not change
            file_contents.append(f'output "{self.outputfile}"\n')
        else:
            file_contents.append(f"output {self.outputfile}\n")

        individual_constraints = all(bool(dct.get("constraints")) for dct in molecules)
        if not individual_constraints and any(bool(dct.get("constraints")) for dct in molecules):
            raise ValueError("Either specify constraints for all molecules or for none of them.")
        has_atoms_constraints = any(bool(dct.get("atoms_constraints")) for dct in molecules)
        if has_atoms_constraints and not individual_constraints:
            raise ValueError("You must use individual_constraints to use atoms_constraints.")
        if individual_constraints:
            for dct in molecules:
                for constraint in dct["constraints"]:
                    if not any(pmc in constraint for pmc in PACKMOL_CONSTRAINTS):
                        raise ValueError(f"Wrong packmol constraint: {constraint}")
                if dct.get("atoms_constraints"):
                    for atoms_constraints_dict in dct.get("atoms_constraints"):
                        for constraint in atoms_constraints_dict["constraints"]:
                            if not any(pmc in constraint for pmc in PACKMOL_CONSTRAINTS):
                                raise ValueError(f"Wrong packmol constraint for specific atoms: {constraint}")
        box_list = None
        if box:
            if individual_constraints:
                raise ValueError("Either specify individual constraints or box.")
            box_list = " ".join(map(str, box))
        elif not individual_constraints:
            # Estimate the total volume of all molecules in cubic Å
            net_volume: float = 0.0
            for dct in molecules:
                mol = dct["coords"] if isinstance(dct["coords"], Molecule) else Molecule.from_file(dct["coords"])

                if mol is None:
                    raise ValueError("Molecule cannot be None.")
                # Pad the calculated length by an amount related to the tolerance parameter
                # the amount to add was determined arbitrarily
                length = (
                    max(np.max(mol.cart_coords[:, i]) - np.min(mol.cart_coords[:, i]) for i in range(3))
                    + self.tolerance
                )
                net_volume += (length**3.0) * float(dct["number"])
            box_length: float = net_volume ** (1 / 3)
            logger.info(f"Auto determined box size is {box_length:.1f} Å per side.")
            box_list = f"0.0 0.0 0.0 {box_length:.1f} {box_length:.1f} {box_length:.1f}"

        for dct in molecules:
            if isinstance(dct["coords"], str | Path):
                mol = Molecule.from_file(dct["coords"])
            elif isinstance(dct["coords"], Molecule):
                mol = dct["coords"]
            else:
                raise TypeError("Molecule is not provided in supported format.")

            fname = f"packmol_{dct['name']}.xyz"
            if mol is None:
                raise ValueError("mol is None")
            mapping[fname] = mol.to(fmt="xyz")
            if " " in str(fname):
                file_contents.append(f"structure {fname!r}")
            else:
                file_contents.append(f"structure {fname}")

            if box_list:
                file_contents.extend(
                    (
                        f"  number {dct['number']}",
                        f"  inside box {box_list}",
                        "end structure\n\n",
                    )
                )
            else:
                atoms_constraints = []
                if dct.get("atoms_constraints"):
                    for atoms_constraints_dict in dct.get("atoms_constraints"):
                        atoms_constraints.append(
                            f"  atoms {' '.join([str(idx + 1) for idx in atoms_constraints_dict['indices']])}"
                        )
                        atoms_constraints.extend(["    " + c.strip() for c in atoms_constraints_dict["constraints"]])
                        atoms_constraints.append("  end atoms")
                file_contents.extend(
                    (
                        f"  number {dct['number']}",
                        *["  " + c.strip() for c in dct["constraints"]],
                        *atoms_constraints,
                        "end structure\n\n",
                    )
                )

        mapping |= {str(self.inputfile): "\n".join(file_contents)}

        return PackmolSet(
            inputs=mapping,
            seed=self.seed,
            inputfile=self.inputfile,
            outputfile=self.outputfile,
            stdoutfile=self.stdoutfile,
            control_params=self.control_params,
            tolerance=self.tolerance,
        )



from __future__ import annotations

import re
from collections import defaultdict
from itertools import islice
from typing import TYPE_CHECKING

import numpy as np

from pymatgen.electronic_structure.core import Spin
from pymatgen.io.lobster.future.core import LobsterInteractionsHolder
from pymatgen.io.lobster.future.versioning import version_processor

if TYPE_CHECKING:
    from typing import ClassVar, Literal

    from numpy.typing import NDArray


class COXXCAR(LobsterInteractionsHolder):
    """Reader for COXXCAR-style files (COOPCAR, COHPCAR, COBICAR).

    Parses LOBSTER's COXXCAR outputs and organizes bond and orbital-resolved interaction data.

    Attributes:
        filename (PathLike): Input file path.
        num_bonds (int): Number of bond interactions reported.
        num_data (int): Number of energy/data points.
        efermi (float): Fermi energy from the file.
        spins (list[Spin]): Present spin channels.
        interactions (list[dict]): Parsed interaction metadata.
        data (np.ndarray): Raw numerical table parsed from the file.
    """

    interactions_regex: ClassVar[str] = (
        r"(?i)([a-z]+\d*(?:\_\d+)?)(?:\[(\-?\d+\s+\-?\d+\s+\-?\d+)\])?(?:\[([^]\s]*)\])?(?:\(([^)]*)\))?"
    )
    coxxcar_type: ClassVar[str]

    @property
    def energies(self) -> NDArray[np.floating]:
        """Return the energy grid.

        Returns:
            NDArray[np.floating]: Energies (first column of self.data).
        """
        return self.data[:, 0]

    def parse_header(self) -> None:
        """Parse the file header and set metadata attributes.

        Args:
            lines (list[str]): Lines of the COXXCAR file.
        """
        data = list(islice(self.iterate_lines(), 2))[1].split()

        self.num_bonds = int(data[0])
        self.num_data = int(data[2])
        self.efermi = float(data[-1])

        if int(data[1]) == 2:
            self.spins = [Spin.up, Spin.down]
        else:
            self.spins = [Spin.up]

    def parse_bonds(self) -> None:
        """Parse the bonds/interactions header block.

        Args:
            lines (list[str]): Lines of the COXXCAR file.
        """
        self.interactions = []

        self.parse_header()

        lines_generator = islice(self.iterate_lines(), 2, self.num_bonds + 2)

        for line in lines_generator:
            if "Average" in line:
                self.interactions.append(
                    {
                        "index": 0,
                        "centers": ["Average"],
                        "orbitals": [None],
                        "cells": [[]],
                        "length": None,
                    }
                )
                continue

            bond_index, bond_data = line.split(":", 1)

            if bond_regex_results := re.search(r"No\.(\d+)", bond_index):
                bond_index = bond_regex_results.group(1)
            else:
                raise ValueError(f"Could not parse bond index from line: {line}")

            length = None

            bond_tmp: dict[str, list] = defaultdict(list)

            centers = bond_data.split("->")
            for center in centers:
                if match := re.search(self.interactions_regex, center):
                    match = match.groups()
                    bond_tmp["centers"].append(match[0])
                else:
                    raise ValueError(f"Could not parse interaction from line: {line}")

                if match[1]:
                    bond_tmp["cells"].append([int(x) for x in match[1].split()])
                else:
                    bond_tmp["cells"].append([])

                bond_tmp["orbitals"].append(match[2])

                if match[3]:
                    length = float(match[3])

            bond = {
                "index": int(bond_index),
                "centers": bond_tmp["centers"],
                "cells": bond_tmp["cells"],
                "orbitals": bond_tmp["orbitals"],
                "length": length,
            }

            self.interactions.append(bond)

    def parse_data(self) -> None:
        """Parse the numerical data block into `self.data` and validate shape.

        Args:
            lines (list[str]): Lines of the COXXCAR file.

        Raises:
            ValueError: If the parsed data array shape does not match the expected shape.
        """
        self.data = np.genfromtxt(
            self.iterate_lines(),
            dtype=np.float64,
            skip_header=self.num_bonds + 2,
            loose=False,
        )

        if self.data.shape != (self.num_data, self.num_bonds * 2 * len(self.spins) + 1):
            raise ValueError(
                f"Data shape {self.data.shape} does not match expected shape "
                f"({self.num_data}, {self.num_bonds * 2 * len(self.spins) + 1})."
            )

        self.process_data_into_interactions()

    def process_data_into_interactions(self) -> None:
        """Populate each interaction dict with 'coxx' and 'icoxx' views.

        Assigns numpy views into `self.data` for each spin channel.
        """
        for i, interaction in enumerate(self.interactions):
            real_indices = self.interaction_indices_to_data_indices_mapping(
                i,
                spins=self.spins,
            )

            interaction["coxx"] = {}
            interaction["icoxx"] = {}

            if len(self.spins) == 1:
                interaction["coxx"][Spin.up] = self.data[:, real_indices[0]]
                interaction["icoxx"][Spin.up] = self.data[:, real_indices[1]]
            else:
                interaction["coxx"][Spin.up] = self.data[:, real_indices[0]]
                interaction["icoxx"][Spin.up] = self.data[:, real_indices[1]]
                interaction["coxx"][Spin.down] = self.data[:, real_indices[2]]
                interaction["icoxx"][Spin.down] = self.data[:, real_indices[3]]

    @version_processor(min_version="5.1")
    def parse_file(self) -> None:
        """Parse the full COXXCAR file (header and data)."""
        self.parse_bonds()
        self.parse_data()

    def get_data_indices_by_properties(
        self,
        indices: list[int] | None = None,
        centers: list[str] | None = None,
        cells: list[list[int]] | None = None,
        orbitals: list[str] | None = None,
        length: tuple[float, float] | None = None,
        spins: list[Literal[1, -1]] | None = None,
        data_type: Literal["coxx", "icoxx"] | None = None,
    ) -> list[int]:
        """Return data-column indices matching the provided interaction properties.

        Args:
            indices (list[int] | None): Interaction indices to filter.
            centers (list[str] | None): Atom centers to filter.
            cells (list[list[int]] | None): Unit cell indices to filter.
            orbitals (list[str] | None): Orbitals to filter.
            length (tuple[float, float] | None): Length range to filter.
            spins (list[Spin] | None): Spins to include.
            data_type (Literal["coxx", "icoxx"] | None): Restrict column type.

        Returns:
            list[int]: Sorted list of data column indices that match the filters.
        """
        return self.interaction_indices_to_data_indices_mapping(
            sorted(
                self.get_interaction_indices_by_properties(
                    indices,
                    centers,
                    cells,
                    orbitals,
                    length,
                )
            ),
            spins=spins or self.spins,
            data_type=data_type,
        )

    def get_data_by_properties(
        self,
        indices: list[int] | None = None,
        centers: list[str] | None = None,
        cells: list[list[int]] | None = None,
        orbitals: list[str] | None = None,
        length: tuple[float, float] | None = None,
        spins: list[Literal[1, -1]] | None = None,
        data_type: Literal["coxx", "icoxx"] | None = None,
    ) -> NDArray[np.floating]:
        """Return the data columns matching the provided interaction properties.

        Args:
            indices (list[int] | None): Interaction indices to filter.
            centers (list[str] | None): Atom centers to filter.
            cells (list[list[int]] | None): Unit cell indices to filter.
            orbitals (list[str] | None): Orbitals to filter.
            length (tuple[float, float] | None): Length range to filter.
            spins (list[Spin] | None): Spins to include.
            data_type (Literal["coxx", "icoxx"] | None): Restrict column type.

        Returns:
            np.ndarray: Array with shape (n_energies, n_selected_columns).
        """
        bond_indices = self.get_interaction_indices_by_properties(indices, centers, cells, orbitals, length)
        spins = spins or self.spins

        return self.data[
            :,
            self.interaction_indices_to_data_indices_mapping(
                bond_indices,
                spins=spins,
                data_type=data_type,
            ),
        ]

    def interaction_indices_to_data_indices_mapping(
        self,
        interaction_indices: int | list[int],
        spins: Literal[1, -1] | list[Literal[1, -1]] | None = None,
        data_type: Literal["coxx", "icoxx"] | None = None,
    ) -> list[int]:
        """Map interaction indices to column indices in `self.data`.

        Args:
            interaction_indices (int | list[int]): Single index or list of interaction indices.
            spins (Spin | list[Spin] | None): Spin(s) to include.
            data_type (Literal["coxx", "icoxx"] | None): Select columns of that type.

        Returns:
            list[int]: Sorted list of integer column indices into `self.data`.

        Raises:
            ValueError: If an invalid Spin is requested.
        """
        if spins is None:
            spins = self.spins

        if spins in (1, -1):
            spins = [spins]
        if isinstance(interaction_indices, int):
            interaction_indices = [interaction_indices]

        if set(spins) - set(self.spins):
            raise ValueError(f"Requested `Spin` {spins} is not valid. Valid `Spin`s are: {self.spins}.")

        index_range = np.arange(0, self.num_bonds * 2 * len(spins) + 1)

        if data_type == "icoxx":
            index_range = index_range[1::2]
        elif data_type == "coxx":
            index_range = index_range[::2]

        real_indices = []
        for bond_index in interaction_indices:
            real_indices.extend([bond_index * 2 + 1, bond_index * 2 + 2])

            if Spin.down in spins:
                real_indices.extend(
                    [
                        (self.num_bonds + bond_index) * 2 + 1,
                        (self.num_bonds + bond_index) * 2 + 2,
                    ]
                )

        real_indices = np.array(real_indices, dtype=int)
        real_indices = np.intersect1d(real_indices, index_range)

        return sorted(real_indices.tolist())


class COBICAR(COXXCAR):
    """Reader for COBICAR.lobster files.

    Attributes:
        coxxcar_type (str): Type of COXXCAR file ("COBICAR").
        is_lcfo (bool): Whether the file is in LCFO format.
    """

    coxxcar_type: ClassVar[str] = "COBICAR"

    is_lcfo: ClassVar[bool] = False

    @classmethod
    def get_default_filename(cls) -> str:
        """Return the default filename for COBICAR."""
        return "COBICAR.LCFO.lobster" if cls.is_lcfo else "COBICAR.lobster"


class COHPCAR(COXXCAR):
    """Reader for COHPCAR.lobster files.

    Attributes:
        coxxcar_type (str): Type of COXXCAR file ("COHPCAR").
        is_lcfo (bool): Whether the file is in LCFO format.
    """

    coxxcar_type: ClassVar[str] = "COHPCAR"

    is_lcfo: ClassVar[bool] = False

    @classmethod
    def get_default_filename(cls) -> str:
        """Return the default filename for COOPCAR."""
        return "COHPCAR.LCFO.lobster" if cls.is_lcfo else "COHPCAR.lobster"


class COOPCAR(COXXCAR):
    """Reader for COOPCAR.lobster files.

    Attributes:
        coxxcar_type (str): Type of COXXCAR file ("COOPCAR").
    """

    coxxcar_type: ClassVar[str] = "COOPCAR"

    @classmethod
    def get_default_filename(cls) -> str:
        """Return the default filename for COOPCAR."""
        return "COOPCAR.lobster"


class COHPCAR_LCFO(COHPCAR):
    """Reader for COHPCAR.LCFO.lobster files.

    Attributes:
        is_lcfo (bool): Always True for LCFO format.
    """

    is_lcfo: ClassVar[bool] = True


class COBICAR_LCFO(COBICAR):
    """Reader for COBICAR.LCFO.lobster files.

    Attributes:
        is_lcfo (bool): Always True for LCFO format.
    """

    is_lcfo: ClassVar[bool] = True



"""Parsers for XTB output files and directories."""

from __future__ import annotations

import logging
import os
import re

from monty.json import MSONable

from pymatgen.core import Molecule
from pymatgen.io.xyz import XYZ

__author__ = "Alex Epstein"
__copyright__ = "Copyright 2020, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Alex Epstein"
__email__ = "aepstein@lbl.gov"
__credits__ = "Sam Blau, Evan Spotte-Smith"


logger = logging.getLogger(__name__)


class CRESTOutput(MSONable):
    """Parse CREST output files."""

    def __init__(self, output_filename, path="."):
        """
        Assumes runtype is iMTD-GC [default].

        Args:
            output_filename (str): Filename to parse
            path (str): Path to directory including output_filename and all
                other xtb output files (crest_best.xyz, etc.)
        """
        self.path = path
        self.filename = output_filename

        self.cmd_options: dict = {}
        self.sorted_structures_energies: list = []
        self.properly_terminated = False
        self._parse_crest_output()

    def _parse_crest_output(self):
        """Parse output file and directory to extract all command line inputs
            and output files.
        Sets the attributes:
            cmd_options: Dict of type {flag: value}
            sorted_structures_energies: n x m x 2 list, for n conformers,
                m rotamers per conformer, and tuple of
                [Molecule, energy]
            properly_terminated: True or False if run properly terminated.
        """
        output_filepath = os.path.join(self.path, self.filename)

        # Get CREST command
        crest_cmd = None
        with open(output_filepath, encoding="utf-8") as xtb_out_file:
            for line in xtb_out_file:
                if "> crest" in line:
                    crest_cmd = line.strip()[8:]
                    break

        split_cmd = crest_cmd.split(" ")

        # Get input file if present
        try:
            self.coord_file = os.path.join(self.path, split_cmd[0])
            self.input_structure = Molecule.from_file(filename=self.coord_file)
        except FileNotFoundError:
            logger.exception(f"Input file {split_cmd[0]} not found")

        # Get CREST input flags
        for idx, entry in enumerate(split_cmd):
            value = None
            if entry and "-" in entry:
                option = entry[1:]
                if idx < len(split_cmd) and "-" not in split_cmd[idx + 1]:
                    value = split_cmd[idx + 1]
                self.cmd_options[option] = value
        # Get input charge for decorating parsed molecules
        chg = 0
        if "chrg" in self.cmd_options:
            str_chg = self.cmd_options["chrg"]
            chg = int(str_chg) if "-" in str_chg else int(str_chg[-1])
        elif "c" in self.cmd_options:
            str_chg = self.cmd_options["c"]
            chg = int(str_chg) if "-" in str_chg else int(str_chg[-1])

        # Check for proper termination
        with open(output_filepath, "rb+") as xtb_out_file:
            xtb_out_file.seek(-2, 2)
            while xtb_out_file.read(1) != b"\n":
                xtb_out_file.seek(-2, 1)
            end_bstring = xtb_out_file.read()
            if b"CREST terminated normally." in end_bstring:
                self.properly_terminated = True

        if self.properly_terminated:
            # Parse for number of conformers and rotamers
            conformer_pattern = re.compile(
                r"\s+\d+\s+(?P<Erel>\d*\.\d*)\s+(?P<Etot>-*\d+\.\d+)\s+"
                r"(?P<weight>-*\d+\.\d+)\s+"
                r"(?P<conformer>-*\d+\.\d+)\s+(?P<set>\d+)\s+(?P<degen>\d+)\s+"
                r"(?P<origin>\w+)\n"
            )
            rotamer_pattern = re.compile(
                r"\s+\d+\s+(?P<Erel>\d*\.\d*)\s+(?P<Etot>-*\d+\.\d+)\s+"
                r"(?P<weight>-*\d+\.\d+)\s+"
                r"(?P<origin>\w+)\n"
            )
            conformer_degeneracies = []
            energies = []
            with open(output_filepath, encoding="utf-8") as xtb_out_file:
                for line in xtb_out_file:
                    conformer_match = conformer_pattern.match(line)
                    rotamer_match = rotamer_pattern.match(line)
                    if conformer_match:
                        conformer_degeneracies.append(int(conformer_match["degen"]))
                        energies.append(conformer_match["Etot"])
                    elif rotamer_match:
                        energies.append(rotamer_match["Etot"])
            # Get final rotamers file and read in all molecules,
            # sorted by conformer type and energy
            if "crest_rotamers.xyz" in os.listdir(self.path):
                final_rotamer_filename = "crest_rotamers.xyz"
            else:
                n_rot_files = []
                for filename in os.listdir(self.path):
                    if "crest_rotamers" in filename:
                        n_rot_file = int(os.path.splitext(filename)[0].split("_")[2])
                        n_rot_files.append(n_rot_file)
                final_rotamer_filename = f"crest_rotamers_{max(n_rot_files)}.xyz" if len(n_rot_files) > 0 else ""

            try:
                rotamers_path = os.path.join(self.path, final_rotamer_filename)
                rotamer_structures = XYZ.from_file(rotamers_path).all_molecules
                for r in rotamer_structures:
                    r.set_charge_and_spin(charge=chg)
                start = 0
                for n, d in enumerate(conformer_degeneracies):
                    self.sorted_structures_energies.append([])
                    idx = 0
                    for idx in range(start, start + d):
                        self.sorted_structures_energies[n].append([rotamer_structures[idx], energies[idx]])
                    start = idx + 1
            except FileNotFoundError:
                logger.exception(f"{final_rotamer_filename} not found, no rotamer list processed")

            # Get lowest energy conformer from 'crest_best.xyz'
            crest_best_path = os.path.join(self.path, "crest_best.xyz")
            try:
                lowest_e_struct = Molecule.from_file(crest_best_path)
                lowest_e_struct.set_charge_and_spin(charge=chg)
                self.lowest_energy_structure = lowest_e_struct
            except FileNotFoundError:
                logger.exception(f"{crest_best_path} not found")

        else:
            crest_best_path = os.path.join(self.path, "crest_best.xyz")
            try:
                lowest_e_struct = Molecule.from_file(crest_best_path)
                lowest_e_struct.set_charge_and_spin(charge=chg)
                self.lowest_energy_structure = lowest_e_struct
            except FileNotFoundError:
                logger.exception(f"{crest_best_path} not found")



"""Package for analyzing elastic tensors and properties."""

from __future__ import annotations

from .elastic import (
    ComplianceTensor,
    ElasticTensor,
    ElasticTensorExpansion,
    NthOrderElasticTensor,
    diff_fit,
    find_eq_stress,
    generate_pseudo,
    get_diff_coeff,
    get_strain_state_dict,
    get_symbol_list,
    raise_if_unphysical,
    subs,
)
from .strain import Deformation, DeformedStructureSet, Strain
from .stress import Stress



"""Classes for parsing and manipulating VASP optical properties calculations."""

from __future__ import annotations

import itertools
from dataclasses import dataclass
from typing import TYPE_CHECKING, overload

import numpy as np
import scipy.constants
import scipy.special
from monty.json import MSONable
from tqdm import tqdm

from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp.outputs import Vasprun, Waveder

if TYPE_CHECKING:
    from typing import Self

    from numpy.typing import ArrayLike, NDArray

    from pymatgen.util.typing import PathLike

__author__ = "Jimmy-Xuan Shen"
__copyright__ = "Copyright 2022, The Materials Project"
__maintainer__ = "Jimmy-Xuan Shen"
__email__ = "jmmshn@gmail.com"

au2ang: float = scipy.constants.physical_constants["atomic unit of length"][0] / 1e-10
ryd2ev: float = scipy.constants.physical_constants["Rydberg constant times hc in eV"][0]
edeps: float = 4 * np.pi * 2 * ryd2ev * au2ang  # from constant.inc in VASP

KB: float = scipy.constants.physical_constants["Boltzmann constant in eV/K"][0]


@dataclass
class DielectricFunctionCalculator(MSONable):
    """Post-process VASP optical properties calculations.

    This objects helps load the different parameters from the vasprun.xml file but allows users to override
    them as needed.

    The standard vasprun.xml from an ``LOPTICS=.True.`` calculation already contains
    the complex frequency dependent dielectric functions.  However you have no way to decompose
    the different contributions.  Since the ``WAVEDER`` file is also written during an optical calculation,
    you can reconstruct the dielectric functions purely in Python and have full control over contribution
    from different bands and k-points.

    VASP's linear optics follow these steps:
        - Calculate the imaginary part
        - Perform symmetry operations (this is not implemented here)
        - Calculate the real part

    Currently, this Calculator only works for ``ISYM=0`` calculations since we cannot guarantee that our
    externally defined symmetry operations are the same as VASP's. This can be fixed by printing the
    symmetry operators into the vasprun.xml file. If this happens in future versions of VASP,
    we can dramatically speed up the calculations here by considering only the irreducible kpoints.
    """

    cder_real: NDArray
    cder_imag: NDArray
    eigs: NDArray
    kweights: NDArray
    nedos: int
    deltae: float
    ismear: int
    sigma: float
    efermi: float
    cshift: float
    ispin: int
    volume: float

    @classmethod
    def from_vasp_objects(cls, vrun: Vasprun, waveder: Waveder) -> Self:
        """Construct a DielectricFunction from Vasprun, Kpoint, and Waveder.

        Args:
            vrun: Vasprun object
            kpoint: Kpoint object
            waveder: Waveder object
        """
        bands = vrun.eigenvalues
        if bands is None:
            raise RuntimeError("eigenvalues cannot be None.")

        sspins = [Spin.up, Spin.down]
        eigs = np.stack([bands[spin] for spin in sspins[: vrun.parameters["ISPIN"]]], axis=2)[..., 0]
        eigs = np.swapaxes(eigs, 0, 1)
        kweights = vrun.actual_kpoints_weights
        nedos = vrun.parameters["NEDOS"]
        deltae = vrun.dielectric[0][1]
        ismear = vrun.parameters["ISMEAR"]
        sigma = vrun.parameters["SIGMA"]
        cshift = vrun.parameters["CSHIFT"]
        efermi = vrun.efermi
        ispin = vrun.parameters["ISPIN"]
        volume = vrun.final_structure.volume
        if vrun.parameters["ISYM"] != 0:
            raise NotImplementedError("ISYM != 0 is not implemented yet")

        if efermi is None:
            raise ValueError("efermi cannot be None.")

        return cls(
            cder_real=waveder.cder_real,
            cder_imag=waveder.cder_imag,
            eigs=eigs,
            kweights=kweights,  # type:ignore[arg-type]
            nedos=nedos,
            deltae=deltae,
            ismear=ismear,
            sigma=sigma,
            efermi=efermi,
            cshift=cshift,
            ispin=ispin,
            volume=volume,
        )

    @classmethod
    def from_directory(cls, directory: PathLike) -> Self:
        """Construct a DielectricFunction from a directory containing
        vasprun.xml and WAVEDER files.
        """

        def _try_reading(dtypes):
            """Return None if failed."""
            for dtype in dtypes:
                try:
                    return Waveder.from_binary(f"{directory}/WAVEDER", data_type=dtype)
                except ValueError as exc:
                    if "reshape" in str(exc):
                        continue
                    raise
            return None

        vrun = Vasprun(f"{directory}/vasprun.xml")
        if "gamma" in vrun.generator["subversion"].lower():
            waveder = _try_reading(["float64", "float32"])  # large one first should give value error
        else:
            waveder = _try_reading(["complex128", "complex64"])
        return cls.from_vasp_objects(vrun, waveder)

    @property
    def cder(self):
        """Complex CDER from WAVEDER."""
        return self.cder_real + self.cder_imag * 1.0j

    def get_epsilon(
        self,
        idir: int,
        jdir: int,
        efermi: float | None = None,
        nedos: int | None = None,
        deltae: float | None = None,
        ismear: int | None = None,
        sigma: float | None = None,
        cshift: float | None = None,
        mask: NDArray | None = None,
    ) -> tuple[NDArray, NDArray]:
        """Compute the frequency dependent dielectric function.

        Args:
            idir: First direction of the dielectric tensor
            jdir: Second direction of the dielectric tensor
            efermi: Fermi energy
            nedos: Number of points in the DOS
            deltae: Energy step in the DOS
            ismear: Smearing method (only has 0:gaussian, >0:Methfessel-Paxton)
            sigma: Smearing width
            cshift: Complex shift used for Kramer-Kronig transformation
            mask: Mask for the bands/kpoint/spin index to include in the calculation
        """

        @overload
        def _use_default(param: int | None, default: int) -> int:
            pass

        @overload
        def _use_default(param: float | None, default: float) -> float:
            pass

        def _use_default(param: float | None, default: float) -> float:
            return param if param is not None else default

        _efermi = _use_default(efermi, self.efermi)
        _nedos = _use_default(nedos, self.nedos)
        _deltae = _use_default(deltae, self.deltae)
        _ismear = _use_default(ismear, self.ismear)
        _sigma = _use_default(sigma, self.sigma)
        _cshift = _use_default(cshift, self.cshift)

        egrid, eps_imag = epsilon_imag(
            cder=self.cder,
            eigs=self.eigs,
            kweights=self.kweights,
            efermi=_efermi,
            nedos=_nedos,
            deltae=_deltae,
            ismear=_ismear,
            sigma=_sigma,
            idir=idir,
            jdir=jdir,
            mask=mask,
        )
        # scaling constant: edeps * np.pi / structure.volume
        eps_in = eps_imag * edeps * np.pi / self.volume
        eps = kramers_kronig(eps_in, nedos=_nedos, deltae=_deltae, cshift=_cshift)
        if idir == jdir:
            eps += 1.0 + 0.0j
        return egrid, eps

    def plot_weighted_transition_data(
        self,
        idir: int,
        jdir: int,
        mask: NDArray | None = None,
        min_val: float = 0.0,
    ):
        """Data for plotting the weight matrix elements as a scatter plot.

        Since the computation of the final spectrum (especially the smearing part)
        is still fairly expensive.  This function can be used to check the values
        of some portion of the spectrum (defined by the mask).
        In a sense, we are looking at the imaginary part of the dielectric function
        before the smearing is applied.

        Args:
            idir: First direction of the dielectric tensor.
            jdir: Second direction of the dielectric tensor.
            mask: Mask to apply to the CDER for the bands/kpoint/spin
                index to include in the calculation
            min_val: Minimum value below this value the matrix element will not be shown.
        """
        cderm = self.cder * mask if mask is not None else self.cder

        norm_kweights = np.array(self.kweights) / np.sum(self.kweights)
        eigs_shifted = self.eigs - self.efermi
        rspin = 3 - cderm.shape[3]

        # Limit the first two indices based on the mask
        try:
            min_band0, max_band0 = np.min(np.where(cderm)[0]), np.max(np.where(cderm)[0])
            min_band1, max_band1 = np.min(np.where(cderm)[1]), np.max(np.where(cderm)[1])

        except ValueError as exc:
            if "zero-size array" in str(exc):
                raise ValueError("No matrix elements found. Check the mask.")
            raise

        x_val = []
        y_val = []
        text = []
        _, _, nk, nspin = cderm.shape[:4]
        iter_idx = [
            range(min_band0, max_band0 + 1),
            range(min_band1, max_band1 + 1),
            range(nk),
            range(nspin),
        ]
        num_ = (max_band0 - min_band0) * (max_band1 - min_band1) * nk * nspin
        for ib, jb, ik, ispin in tqdm(itertools.product(*iter_idx), total=num_):
            fermi_w_i = step_func((eigs_shifted[ib, ik, ispin]) / self.sigma, self.ismear)
            fermi_w_j = step_func((eigs_shifted[jb, ik, ispin]) / self.sigma, self.ismear)
            weight = (fermi_w_j - fermi_w_i) * rspin * norm_kweights[ik]
            A = cderm[ib, jb, ik, ispin, idir] * np.conjugate(cderm[ib, jb, ik, ispin, jdir])
            decel = self.eigs[jb, ik, ispin] - self.eigs[ib, ik, ispin]
            matrix_el = np.abs(A) * float(weight)  # can have negative weight due to fermi function
            if matrix_el > min_val:
                x_val.append(decel)
                y_val.append(matrix_el)
                text.append(f"s:{ispin}, k:{ik}, {ib} -> {jb} ({decel:.2f})")
        return x_val, y_val, text


def delta_methfessel_paxton(x: NDArray, n: int) -> NDArray:
    """
    D_n (x) = exp -x^2 * sum_i=0^n A_i H_2i(x)
    where H is a Hermite polynomial and
    A_i = (-1)^i / ( i! 4^i sqrt(pi) ).
    """
    ii = np.arange(0, n + 1)
    A = (-1) ** ii / (scipy.special.factorial(ii) * 4**ii * np.sqrt(np.pi))
    H = scipy.special.eval_hermite(ii * 2, np.tile(x, (len(ii), 1)).T)
    return np.exp(-(x * x)) * np.dot(A, H.T)


def step_methfessel_paxton(x: NDArray, n: int) -> NDArray:
    """
    S_n (x) = (1 + erf x)/2 - exp -x^2 * sum_i=1^n A_i H_{2i-1}(x)
    where H is a Hermite polynomial and
    A_i = (-1)^i / ( i! 4^i sqrt(pi) ).
    """
    ii = np.arange(1, n + 1)
    A = (-1) ** ii / (scipy.special.factorial(ii) * 4**ii * np.sqrt(np.pi))
    H = scipy.special.eval_hermite(ii * 2 - 1, np.tile(x, (len(ii), 1)).T)
    return (1.0 + scipy.special.erf(x)) / 2.0 - np.exp(-(x * x)) * np.dot(A, H.T)


def delta_func(x: NDArray, ismear: int) -> NDArray:
    """Replication of VASP's delta function."""
    if ismear < -1:
        raise ValueError("Delta function not implemented for ismear < -1")
    if ismear == -1:
        return step_func(x, -1) * (1 - step_func(x, -1))
    if ismear == 0:
        return np.exp(-(x * x)) / np.sqrt(np.pi)
    return delta_methfessel_paxton(x, ismear)


def step_func(x: NDArray, ismear: int) -> NDArray:
    """Replication of VASP's step function."""
    if ismear < -1:
        raise ValueError("Delta function not implemented for ismear < -1")
    if ismear == -1:
        return 1 / (1.0 + np.exp(-x))
    if ismear == 0:
        return 0.5 + 0.5 * scipy.special.erf(x)
    return step_methfessel_paxton(x, ismear)


def get_delta(x0: float, sigma: float, nx: int, dx: float, ismear: int = 3) -> NDArray:
    """Get the smeared delta function to be added to form the spectrum.

    This replaces the `SLOT` function from VASP. Uses finite differences instead of
    evaluating the delta function since the step function is more likely to have analytic form.

    Args:
        x0: The center of the dielectric function.
        sigma: The width of the smearing
        nx: The number of grid points in the output grid.
        dx: The gridspacing of the output grid.
        ismear: The smearing parameter used by the ``step_func``.

    Returns:
        np.ndarray: Array of size `nx` with delta function on the desired outputgrid.
    """
    xgrid = np.linspace(0, nx * dx, nx, endpoint=False)
    xgrid -= x0
    x_scaled = (xgrid + (dx / 2)) / sigma
    sfun = step_func(x_scaled, ismear)
    dfun = np.zeros_like(xgrid)
    dfun[1:] = (sfun[1:] - sfun[:-1]) / dx
    return dfun


def get_step(x0: float, sigma: float, nx: int, dx: float, ismear: int) -> NDArray:
    """Get the smeared step function to be added to form the spectrum.

    This replaces the `SLOT` function from VASP.

    Args:
        x0: The center of the dielectric function.
        sigma: The width of the smearing
        nx: The number of grid points in the output grid.
        dx: The gridspacing of the output grid.
        ismear: The smearing parameter used by the ``step_func``.

    Returns:
        np.ndarray: Array of size `nx` with step function on the desired outputgrid.
    """
    xgrid = np.linspace(0, nx * dx, nx, endpoint=False)
    xgrid -= x0
    x_scaled = (xgrid + (dx / 2)) / sigma
    return step_func(x_scaled, ismear)


def epsilon_imag(
    cder: NDArray,
    eigs: NDArray,
    kweights: ArrayLike,
    efermi: float,
    nedos: int,
    deltae: float,
    ismear: int,
    sigma: float,
    idir: int,
    jdir: int,
    mask: NDArray | None = None,
) -> tuple[NDArray, NDArray]:
    """Replicate the EPSILON_IMAG function of VASP.

    Args:
        cder: The data written to the WAVEDER (nbands, nbands, nkpoints, nspin, diri, dirj)
        eigs: The eigenvalues (nbands, nkpoints, nspin)
        kweights: The kpoint weights (nkpoints)
        efermi: The fermi energy
        nedos: The sampling of the energy values
        deltae: The energy grid spacing
        ismear: The smearing parameter used by the ``step_func``.
        sigma: The width of the smearing
        idir: The first direction of the dielectric tensor
        jdir: The second direction of the dielectric tensor
        mask: Mask for the bands/kpoint/spin index to include in the calculation

    Returns:
        np.ndarray: Array of size `nedos` with the imaginary part of the dielectric function.
    """
    norm_kweights = np.array(kweights) / np.sum(kweights)
    egrid = np.linspace(0, nedos * deltae, nedos, endpoint=False)
    eigs_shifted = eigs - efermi
    # np.subtract.outer results in a matrix of shape (nband, nband)
    rspin = 3 - cder.shape[3]

    # For the transition between two bands at one kpoint the contributions is:
    #  (fermi[band_i] - fermi[band_j]) * rspin * normalized_kpoint_weight
    cderm = cder * mask if mask is not None else cder

    # min_band0, max_band0 = np.min(np.where(cderm)[0]), np.max(np.where(cderm)[0])
    # min_band1, max_band1 = np.min(np.where(cderm)[1]), np.max(np.where(cderm)[1])
    # limit the first two indices based on the mask
    try:
        min_band0, max_band0 = np.min(np.where(cderm)[0]), np.max(np.where(cderm)[0])
        min_band1, max_band1 = np.min(np.where(cderm)[1]), np.max(np.where(cderm)[1])
    except ValueError as exc:
        if "zero-size array" in str(exc):
            return egrid, np.zeros_like(egrid, dtype=np.complex128)
        raise
    _, _, nk, nspin = cderm.shape[:4]
    iter_idx = [
        range(min_band0, max_band0 + 1),
        range(min_band1, max_band1 + 1),
        range(nk),
        range(nspin),
    ]
    num_ = (max_band0 - min_band0) * (max_band1 - min_band1) * nk * nspin
    epsdd = np.zeros_like(egrid, dtype=np.complex128)
    for ib, jb, ik, ispin in tqdm(itertools.product(*iter_idx), total=num_):
        fermi_w_i = step_func((eigs_shifted[ib, ik, ispin]) / sigma, ismear)
        fermi_w_j = step_func((eigs_shifted[jb, ik, ispin]) / sigma, ismear)
        weight = (fermi_w_j - fermi_w_i) * rspin * norm_kweights[ik]
        decel = eigs[jb, ik, ispin] - eigs[ib, ik, ispin]
        A = cderm[ib, jb, ik, ispin, idir] * np.conjugate(cderm[ib, jb, ik, ispin, jdir])
        # Reproduce the `SLOT` function calls in VASP:
        # CALL SLOT( REAL(DECEL,q), ISMEAR, SIGMA, NEDOS, DELTAE,  WEIGHT*A*CONST, EPSDD)
        # The conjugate part is not needed since we are running over all pairs of ib, jb
        # vasp just does the conjugate trick to save loop time
        smeared = get_delta(x0=decel, sigma=sigma, nx=nedos, dx=deltae, ismear=ismear) * weight * A
        epsdd += smeared
    return egrid, epsdd


def kramers_kronig(
    eps: NDArray,
    nedos: int,
    deltae: float,
    cshift: float = 0.1,
) -> NDArray:
    """Perform the Kramers-Kronig transformation.

    Perform the Kramers-Kronig transformation exactly as VASP does it.
    The input eps should be complex and the imaginary part of the dielectric function
    should be stored as the real part of the complex input array.
    The output should be the complex dielectric function.

    Args:
        eps: The dielectric function with the imaginary part stored as the real part
            and nothing in the imaginary part.
        nedos: The sampling of the energy values
        deltae: The energy grid spacing
        cshift: The shift of the imaginary part of the dielectric function.

    Returns:
        np.ndarray: Array of size `nedos` with the complex dielectric function.
    """
    egrid = np.linspace(0, deltae * nedos, nedos)
    csfhit = cshift * 1.0j
    cdiff = np.subtract.outer(egrid, egrid) + csfhit
    csum = np.add.outer(egrid, egrid) + csfhit
    vals = -0.5 * ((eps / cdiff) - (np.conj(eps) / csum))
    return np.sum(vals, axis=1) * 2 / np.pi * deltae



"""This module defines filters for Transmuter object."""

from __future__ import annotations

import abc
import math
from collections import defaultdict
from typing import TYPE_CHECKING

from monty.json import MSONable

from pymatgen.core import get_el_sp
from pymatgen.core.structure_matcher import ElementComparator, StructureMatcher
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

if TYPE_CHECKING:
    from typing import Self

    from pymatgen.core import IStructure, Structure
    from pymatgen.util.typing import SpeciesLike


class AbstractStructureFilter(MSONable, abc.ABC):
    """Structures that return True when passed to the test() method are retained during
    transmutation. Those that return False are removed.
    """

    @abc.abstractmethod
    def test(self, structure: Structure | IStructure):
        """Structures that return true are kept in the Transmuter object during filtering.

        Args:
            structure (Structure): Input structure to test

        Returns:
            bool: True if structure passes filter.
        """


class ContainsSpecieFilter(AbstractStructureFilter):
    """Filter for structures containing certain elements or species.
    By default compares by atomic number.
    """

    def __init__(self, species: SpeciesLike, strict_compare: bool = False, AND: bool = True, exclude: bool = False):
        """
        Args:
            species (list[SpeciesLike]): species to look for
            AND: whether all species must be present to pass (or fail) filter.
            strict_compare: if true, compares objects by specie or element
                object if false, compares atomic number
            exclude: If true, returns false for any structures with the specie
                (excludes them from the Transmuter).
        """
        self._species = list(map(get_el_sp, species))  # type:ignore[arg-type]
        self._strict = strict_compare
        self._AND = AND
        self._exclude = exclude

    def test(self, structure: Structure | IStructure):
        """True if structure does not contain specified species."""
        # set up lists to compare
        if not self._strict:
            # compare by atomic number
            filter_set = {sp.Z for sp in self._species}
            structure_set = {sp.Z for sp in structure.elements}
        else:
            # compare by specie or element object
            filter_set = set(self._species)
            structure_set = set(structure.elements)

        if self._AND and filter_set <= structure_set:
            # return true if we aren't excluding since all are in structure
            return not self._exclude
        if not self._AND and filter_set & structure_set:
            # return true if we aren't excluding since one is in structure
            return not self._exclude
        # return false if we aren't excluding otherwise
        return self._exclude

    def __repr__(self):
        return "\n".join(
            [
                "ContainsSpecieFilter with parameters:",
                f"species = {self._species}",
                f"strict_compare = {self._strict}",
                f"AND = {self._AND}",
                f"exclude = {self._exclude}",
            ]
        )

    def as_dict(self) -> dict:
        """Get MSONable dict."""
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "init_args": {
                "species": [str(sp) for sp in self._species],
                "strict_compare": self._strict,
                "AND": self._AND,
                "exclude": self._exclude,
            },
        }

    @classmethod
    def from_dict(cls, dct: dict) -> Self:
        """
        Args:
            dct (dict): Dict representation.

        Returns:
            Filter
        """
        return cls(**dct["init_args"])


class SpecieProximityFilter(AbstractStructureFilter):
    """This filter removes structures that have certain species that are too close together."""

    def __init__(self, specie_and_min_dist_dict):
        """
        Args:
            specie_and_min_dist_dict (dict): A species string to float mapping. For
                example, {"Na+": 1} means that all Na+ ions must be at least 1
                Angstrom away from each other. Multiple species criteria can be
                applied. Note that the testing is done based on the actual object
                . If you have a structure with Element, you must use {"Na":1}
                instead to filter based on Element and not Species.
        """
        self.specie_and_min_dist = {get_el_sp(k): v for k, v in specie_and_min_dist_dict.items()}

    def test(self, structure: Structure):
        """True if structure does not contain species within specified distances."""
        all_species = set(self.specie_and_min_dist)
        for site in structure:
            species = set(site.species)
            if sp_to_test := species.intersection(all_species):
                max_r = max(self.specie_and_min_dist[sp] for sp in sp_to_test)
                neighbors = structure.get_neighbors(site, max_r)
                for sp in sp_to_test:
                    for nn_site, dist, *_ in neighbors:
                        if sp in nn_site.species and dist < self.specie_and_min_dist[sp]:
                            return False
        return True

    def as_dict(self):
        """Get MSONable dict."""
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "init_args": {"specie_and_min_dist_dict": {str(sp): v for sp, v in self.specie_and_min_dist.items()}},
        }

    @classmethod
    def from_dict(cls, dct: dict) -> Self:
        """
        Args:
            dct (dict): Dict representation.

        Returns:
            SpecieProximityFilter
        """
        return cls(**dct["init_args"])


class RemoveDuplicatesFilter(AbstractStructureFilter):
    """This filter removes exact duplicate structures from the transmuter."""

    def __init__(
        self,
        structure_matcher: dict | StructureMatcher | None = None,
        symprec: float | None = None,
    ) -> None:
        """Remove duplicate structures based on the structure matcher
        and symmetry (if symprec is given).

        Args:
            structure_matcher (dict | StructureMatcher, optional): Provides a structure matcher to be used for
                structure comparison.
            symprec (float, optional): The precision in the symmetry finder algorithm if None (
                default value), no symmetry check is performed and only the
                structure matcher is used. A recommended value is 1e-5.
        """
        self.symprec = symprec
        self.structure_list: dict[str, list[Structure]] = defaultdict(list)
        if not isinstance(structure_matcher, dict | StructureMatcher | type(None)):
            raise TypeError(f"{structure_matcher=} must be a dict, StructureMatcher or None")
        if isinstance(structure_matcher, dict):
            self.structure_matcher = StructureMatcher.from_dict(structure_matcher)
        else:
            self.structure_matcher = structure_matcher or StructureMatcher(comparator=ElementComparator())

    def test(self, structure: Structure) -> bool:
        """
        Args:
            structure (Structure): Input structure to test.

        Returns:
            bool: True if structure is not in list.
        """
        hash_comp = self.structure_matcher._comparator.get_hash(structure.composition)
        if not self.structure_list[hash_comp]:
            self.structure_list[hash_comp].append(structure)
            return True

        def get_spg_num(struct: Structure) -> int:
            finder = SpacegroupAnalyzer(struct, symprec=self.symprec)  # type:ignore[arg-type]
            return finder.get_space_group_number()

        for struct in self.structure_list[hash_comp]:
            if (self.symprec is None or get_spg_num(struct) == get_spg_num(structure)) and self.structure_matcher.fit(
                struct, structure
            ):
                return False

        self.structure_list[hash_comp].append(structure)
        return True


class RemoveExistingFilter(AbstractStructureFilter):
    """This filter removes structures existing in a given list from the transmuter."""

    def __init__(
        self,
        existing_structures: list[Structure],
        structure_matcher: dict | StructureMatcher | None = None,
        symprec: float | None = None,
    ) -> None:
        """Remove existing structures based on the structure matcher
        and symmetry (if symprec is given).

        Args:
            existing_structures (list[Structure]): Existing structures to compare with.
            structure_matcher (dict | StructureMatcher, optional): Will be used for
                structure comparison.
            symprec (float | None): The precision in the symmetry finder algorithm.
                If None (default value), no symmetry check is performed and only the
                structure matcher is used. A recommended value is 1e-5.
        """
        self.symprec = symprec
        self.structure_list: list = []
        self.existing_structures = existing_structures
        if isinstance(structure_matcher, dict):
            self.structure_matcher = StructureMatcher.from_dict(structure_matcher)
        else:
            self.structure_matcher = structure_matcher or StructureMatcher(comparator=ElementComparator())

    def test(self, structure: Structure):
        """True if structure is not in existing list."""

        def get_sg(s):
            finder = SpacegroupAnalyzer(s, symprec=self.symprec)
            return finder.get_space_group_number()

        for struct in self.existing_structures:
            if (
                (
                    self.structure_matcher._comparator.get_hash(structure.composition)
                    == self.structure_matcher._comparator.get_hash(struct.composition)
                    and self.symprec is None
                )
                or get_sg(struct) == get_sg(structure)
            ) and self.structure_matcher.fit(struct, structure):
                return False

        self.structure_list.append(structure)
        return True

    def as_dict(self):
        """Get MSONable dict."""
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "init_args": {"structure_matcher": self.structure_matcher.as_dict()},
        }


class ChargeBalanceFilter(AbstractStructureFilter):
    """This filter removes structures that are not charge balanced from the
    transmuter. This only works if the structure is oxidation state
    decorated, as structures with only elemental sites are automatically
    assumed to have net charge of 0.
    """

    def __init__(self):
        """No args required."""

    def test(self, structure: Structure):
        """True if structure is neutral."""
        return math.isclose(structure.charge, 0.0)


class SpeciesMaxDistFilter(AbstractStructureFilter):
    """This filter removes structures that do have two particular species that are
    not nearest neighbors by a predefined max_dist. For instance, if you are
    analyzing Li battery materials, you would expect that each Li+ would be
    nearest neighbor to lower oxidation state transition metal for
    electrostatic reasons. This only works if the structure is oxidation state
    decorated, as structures with only elemental sites are automatically
    assumed to have net charge of 0.
    """

    def __init__(self, sp1, sp2, max_dist):
        """
        Args:
            sp1 (Species): First specie
            sp2 (Species): Second specie
            max_dist (float): Maximum distance between species.
        """
        self.sp1 = get_el_sp(sp1)
        self.sp2 = get_el_sp(sp2)
        self.max_dist = max_dist

    def test(self, structure: Structure):
        """True if structure contains the two species but their distance is greater than max_dist."""
        sp1_indices = [idx for idx, site in enumerate(structure) if site.specie == self.sp1]
        sp2_indices = [idx for idx, site in enumerate(structure) if site.specie == self.sp2]
        frac_coords1 = structure.frac_coords[sp1_indices, :]
        frac_coords2 = structure.frac_coords[sp2_indices, :]
        lattice = structure.lattice
        dists = lattice.get_all_distances(frac_coords1, frac_coords2)
        return all(any(row) for row in dists < self.max_dist)



"""Classes for reading/manipulating/writing exciting input files."""

from __future__ import annotations

import itertools
import warnings
import xml.etree.ElementTree as ET
from typing import TYPE_CHECKING

import numpy as np
import scipy.constants as const
from monty.io import zopen
from monty.json import MSONable

from pymatgen.core import Element, Lattice, Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.bandstructure import HighSymmKpath

if TYPE_CHECKING:
    from typing import ClassVar, Literal, Self

    from numpy.typing import ArrayLike, NDArray

    from pymatgen.core.structure import IStructure
    from pymatgen.util.typing import PathLike

__author__ = "Christian Vorwerk"
__copyright__ = "Copyright 2016"
__version__ = "1.0"
__maintainer__ = "Christian Vorwerk"
__email__ = "vorwerk@physik.hu-berlin.de"
__status__ = "Development"
__date__ = "Nov 28, 2016"


class ExcitingInput(MSONable):
    """
    Object for representing the data stored in the structure part of the
    exciting input.

    Attributes:
        structure (Structure): Associated Structure.
        title (str): Optional title string.
        lockxyz (Nx3 NDArray of booleans): Lockxyz attribute for each site if available.
    """

    # Conversion factor between Bohr radius and Angstrom
    bohr2ang: ClassVar[float] = const.value("Bohr radius") / const.value("Angstrom star")

    def __init__(
        self,
        structure: Structure | IStructure,
        title: str | None = None,
        lockxyz: ArrayLike | None = None,
    ) -> None:
        """
        Args:
            structure (Structure): Structure object.
            title (str): Optional title for exciting input. Defaults to unit
                cell formula of structure. Defaults to None.
            lockxyz (Nx3 array): bool values for selective dynamics,
                where N is number of sites. Defaults to None.
        """
        if structure.is_ordered:
            site_properties = {}
            if lockxyz:
                site_properties["selective_dynamics"] = np.asarray(lockxyz)
            self.structure = structure.copy(site_properties=site_properties)
            self.title = structure.formula if title is None else title
        else:
            raise ValueError("Structure with partial occupancies cannot be converted into exciting input!")

    @property
    def lockxyz(self) -> NDArray | None:
        """Selective dynamics site properties."""
        return self.structure.site_properties.get("selective_dynamics")  # type:ignore[return-value]

    @lockxyz.setter
    def lockxyz(self, lockxyz: ArrayLike):
        self.structure.add_site_property("selective_dynamics", np.array(lockxyz))

    @classmethod
    def from_str(cls, data: str) -> Self:
        """Reads the exciting input from a string."""
        root: ET.Element = ET.fromstring(data)
        struct = root.find("structure")
        if struct is None:
            raise ValueError("No structure found in input file!")

        species_node = struct.iter("species")
        elements = []
        positions = []
        vectors = []
        lockxyz = []
        # get title
        _title = root.find("title")
        if _title is None:
            raise ValueError("title cannot be None.")
        title_in = str(_title.text)
        # Read elements and coordinates
        for nodes in species_node:
            _speciesfile = nodes.get("speciesfile")
            if _speciesfile is None:
                raise ValueError("speciesfile cannot be None.")
            symbol = _speciesfile.split(".")[0]
            if len(symbol.split("_")) == 2:
                symbol = symbol.split("_")[0]
            if Element.is_valid_symbol(symbol):
                # Try to recognize the element symbol
                element = symbol
            else:
                raise ValueError("Unknown element!")

            for atom in nodes.iter("atom"):
                _coord = atom.get("coord")
                if _coord is None:
                    raise ValueError("coordinate cannot be None.")
                x, y, z = _coord.split()
                positions.append([float(x), float(y), float(z)])
                elements.append(element)
                # Obtain lockxyz for each atom
                if atom.get("lockxyz") is None:
                    lockxyz.append([False, False, False])
                else:
                    lxyz = []

                    _lockxyz = atom.get("lockxyz")
                    if _lockxyz is None:
                        raise ValueError("lockxyz cannot be None.")
                    for line in _lockxyz.split():
                        if line in ("True", "true"):
                            lxyz.append(True)
                        else:
                            lxyz.append(False)
                    lockxyz.append(lxyz)

        # check the atomic positions type
        cartesian = False
        if struct.attrib.get("cartesian"):
            cartesian = True
            for p, j in itertools.product(positions, range(3)):
                p[j] *= ExcitingInput.bohr2ang

        _crystal = struct.find("crystal")
        if _crystal is None:
            raise ValueError("crystal cannot be None.")

        # get the scale attribute
        scale_in = _crystal.get("scale")
        scale = float(scale_in) * ExcitingInput.bohr2ang if scale_in else ExcitingInput.bohr2ang

        # get the stretch attribute
        stretch_in = _crystal.get("stretch")
        stretch = np.array([float(a) for a in stretch_in]) if stretch_in else np.array([1.0, 1.0, 1.0])

        # get basis vectors and scale them accordingly
        basisnode = _crystal.iter("basevect")
        for vect in basisnode:
            if vect.text is None:
                raise ValueError("vect.text cannot be None.")
            x, y, z = vect.text.split()
            vectors.append(
                [
                    float(x) * stretch[0] * scale,
                    float(y) * stretch[1] * scale,
                    float(z) * stretch[2] * scale,
                ]
            )
        # create lattice and structure object
        lattice_in = Lattice(vectors)
        structure_in = Structure(lattice_in, elements, positions, coords_are_cartesian=cartesian)

        return cls(structure_in, title_in, lockxyz)

    @classmethod
    def from_file(cls, filename: PathLike) -> Self:
        """
        Args:
            filename: Filename.

        Returns:
            ExcitingInput
        """
        with zopen(filename, mode="rt", encoding="utf-8") as file:
            data = file.read().replace("\n", "")  # type:ignore[arg-type]
        return cls.from_str(data)  # type:ignore[arg-type]

    def write_etree(
        self,
        celltype: Literal["unchanged", "conventional", "primitive"],
        cartesian: bool = False,
        bandstr: bool = False,
        symprec: float = 0.4,
        angle_tolerance: float = 5,
        **kwargs,
    ) -> ET.Element:
        """Convert the exciting input parameters to an XML Element object.

        Args:
            celltype (str): Choice of unit cell. Can be either the unit cell
                from self.structure ("unchanged"), the conventional cell
                ("conventional"), or the primitive unit cell ("primitive").
            cartesian (bool): Whether the atomic positions are provided in
                Cartesian or unit-cell coordinates. Default is False.
            bandstr (bool): Whether the bandstructure path along the
                HighSymmKpath is included in the input file. Only supported if the
                celltype is set to "primitive". Default is False.
            symprec (float): Tolerance for the symmetry finding. Default is 0.4.
            angle_tolerance (float): Angle tolerance for the symmetry finding.
                Default is 5.
            **kwargs: Additional parameters for the input file.

        Returns:
            ET.Element containing the input XML structure
        """
        root = ET.Element("input")
        root.set(
            "{http://www.w3.org/2001/XMLSchema-instance}noNamespaceSchemaLocation",
            "http://xml.exciting-code.org/excitinginput.xsd",
        )
        title = ET.SubElement(root, "title")
        title.text = self.title
        if cartesian:
            structure = ET.SubElement(root, "structure", cartesian="true", speciespath="./")
        else:
            structure = ET.SubElement(root, "structure", speciespath="./")

        crystal = ET.SubElement(structure, "crystal")
        # set scale such that lattice vector can be given in Angstrom
        ang2bohr = const.value("Angstrom star") / const.value("Bohr radius")
        crystal.set("scale", str(ang2bohr))
        # determine which structure to use
        finder = SpacegroupAnalyzer(self.structure, symprec=symprec, angle_tolerance=angle_tolerance)
        if celltype == "primitive":
            new_struct = finder.get_primitive_standard_structure(international_monoclinic=False)
        elif celltype == "conventional":
            new_struct = finder.get_conventional_standard_structure(international_monoclinic=False)
        elif celltype == "unchanged":
            new_struct = self.structure
        else:
            raise ValueError("Type of unit cell not recognized!")

        # write lattice
        basis = new_struct.lattice.matrix
        for idx in range(3):
            base_vec = ET.SubElement(crystal, "basevect")
            base_vec.text = f"{basis[idx][0]:16.8f} {basis[idx][1]:16.8f} {basis[idx][2]:16.8f}"
        # write atomic positions for each species
        index = 0
        for elem in sorted(new_struct.types_of_species, key=lambda el: el.X):
            species = ET.SubElement(structure, "species", speciesfile=f"{elem.symbol}.xml")
            sites = new_struct.indices_from_symbol(elem.symbol)

            for j in sites:
                fc = new_struct[j].frac_coords
                coord = f"{fc[0]:16.8f} {fc[1]:16.8f} {fc[2]:16.8f}"
                # obtain Cartesian coords from fractional ones if needed
                if cartesian:
                    coord2 = []
                    for k in range(3):
                        inter = (
                            new_struct[j].frac_coords[k] * basis[0][k]
                            + new_struct[j].frac_coords[k] * basis[1][k]
                            + new_struct[j].frac_coords[k] * basis[2][k]
                        ) * ang2bohr
                        coord2.append(inter)
                    coord = f"{coord2[0]:16.8f} {coord2[1]:16.8f} {coord2[2]:16.8f}"

                # write atomic positions
                index += 1
                _ = ET.SubElement(species, "atom", coord=coord)

        # write bandstructure if needed
        if bandstr:
            if celltype != "primitive":
                raise ValueError("Bandstructure is only implemented for the standard primitive unit cell!")

            kpath = HighSymmKpath(new_struct, symprec=symprec, angle_tolerance=angle_tolerance)
            prop = ET.SubElement(root, "properties")
            band_struct = ET.SubElement(prop, "bandstructure")
            for idx in range(len(kpath.kpath["path"])):
                plot = ET.SubElement(band_struct, "plot1d")
                path = ET.SubElement(plot, "path", steps="100")
                for j in range(len(kpath.kpath["path"][idx])):
                    symbol = kpath.kpath["path"][idx][j]
                    coords = kpath.kpath["kpoints"][symbol]
                    coord = f"{coords[0]:16.8f} {coords[1]:16.8f} {coords[2]:16.8f}"
                    symbol_map = {
                        "\\Gamma": "GAMMA",
                        "\\Sigma": "SIGMA",
                        "\\Delta": "DELTA",
                        "\\Lambda": "LAMBDA",
                    }
                    symbol = symbol_map.get(symbol, symbol)
                    _ = ET.SubElement(path, "point", coord=coord, label=symbol)

        # write extra parameters from kwargs if provided
        self._dicttoxml(kwargs, root)

        return root

    def write_string(
        self,
        celltype: Literal["unchanged", "conventional", "primitive"],
        cartesian: bool = False,
        bandstr: bool = False,
        symprec: float = 0.4,
        angle_tolerance: float = 5,
        **kwargs,
    ) -> str:
        """Convert exciting input to a string.

        Args:
            celltype (str): Choice of unit cell. Can be either the unit cell
                from self.structure ("unchanged"), the conventional cell
                ("conventional"), or the primitive unit cell ("primitive").
            cartesian (bool): Whether the atomic positions are provided in
                Cartesian or unit-cell coordinates. Default is False.
            bandstr (bool): Whether the bandstructure path along the
                HighSymmKpath is included in the input file. Only supported if the
                celltype is set to "primitive". Default is False.
            symprec (float): Tolerance for the symmetry finding. Default is 0.4.
            angle_tolerance (float): Angle tolerance for the symmetry finding.
                Default is 5.
            **kwargs: Additional parameters for the input file.

        Returns:
            str
        """
        try:
            root = self.write_etree(celltype, cartesian, bandstr, symprec, angle_tolerance, **kwargs)
            self._indent(root)
            # output should be a string not a bytes object
            string = ET.tostring(root).decode("UTF-8")

        except Exception:
            raise ValueError("Incorrect celltype!")

        return string

    def write_file(
        self,
        celltype: Literal["unchanged", "conventional", "primitive"],
        filename: PathLike,
        cartesian: bool = False,
        bandstr: bool = False,
        symprec: float = 0.4,
        angle_tolerance: float = 5,
        **kwargs,
    ) -> None:
        """Write exciting input file.

        Args:
            celltype (str): Choice of unit cell. Can be either the unit cell
                from self.structure ("unchanged"), the conventional cell
                ("conventional"), or the primitive unit cell ("primitive").
            filename (PathLike): Filename for exciting input.
            cartesian (bool): Whether the atomic positions are provided in
                Cartesian or unit-cell coordinates. Default is False.
            bandstr (bool): Whether the bandstructure path along the
                HighSymmKpath is included in the input file. Only supported if the
                celltype is set to "primitive". Default is False.
            symprec (float): Tolerance for the symmetry finding. Default is 0.4.
            angle_tolerance (float): Angle tolerance for the symmetry finding.
                Default is 5.
            **kwargs: Additional parameters for the input file.
        """
        try:
            root = self.write_etree(celltype, cartesian, bandstr, symprec, angle_tolerance, **kwargs)
            self._indent(root)
            tree = ET.ElementTree(root)
            tree.write(filename)
        except Exception:
            raise ValueError("Incorrect celltype!")

    @staticmethod
    def _indent(elem: ET.Element, level: int = 0) -> None:
        """
        Helper method to indent elements, as missing PrettyPrint option
        in the current version of xml.etree.cElementTree.

        Args:
            elem (ET.Element): The Element to process.
            level (int): The indentation level.
        """
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = f"{i}  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for el in elem:
                ExcitingInput._indent(el, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        elif level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i

    def _dicttoxml(self, paramdict_: dict, element: ET.Element) -> None:
        for key, value in paramdict_.items():
            if isinstance(value, str) and key == "text()":
                element.text = value

            elif isinstance(value, str):
                element.attrib[key] = value

            elif isinstance(value, list):
                for item in value:
                    self._dicttoxml(item, ET.SubElement(element, key))

            elif isinstance(value, dict):
                if element.findall(key) == []:
                    self._dicttoxml(value, ET.SubElement(element, key))
                else:
                    self._dicttoxml(value, element.findall(key)[0])

            else:
                warnings.warn(f"cannot deal with {key} = {value}", stacklevel=2)



"""This module implements a MolecularOrbital class to represent band character in
solids. Useful for predicting PDOS character from structural information.
"""

from __future__ import annotations

from itertools import chain, combinations
from typing import TYPE_CHECKING

from pymatgen.core import Element
from pymatgen.core.composition import Composition

if TYPE_CHECKING:
    from typing import Any


class MolecularOrbitals:
    """Represents the character of bands in a solid. The input is a chemical
    formula, since no structural characteristics are taken into account.

    The band character of a crystal emerges from the atomic orbitals of the
    constituent ions, hybridization/covalent bonds, and the spin-orbit
    interaction (ex: Fe2O3). Right now the orbitals are only built from
    the uncharged atomic species. Functionality can be improved by:
    1) calculate charged ion orbital energies
    2) incorporate the coordination environment to account for covalent bonds

    The atomic orbital energies are stored in pymatgen.core.periodic_table.JSON

    MOs = MolecularOrbitals('SrTiO3')
    MOs.band_edges
    # gives {'HOMO':['O','2p',-0.338381], 'LUMO':['Ti','3d',-0.17001], 'metal':False}
    """

    def __init__(self, formula: str) -> None:
        """
        Args:
            formula (str): Chemical formula. Must have integer subscripts. Ex: 'SrTiO3'.

        Attributes:
            composition: the composition as a dictionary. Ex: {'Sr': 1, 'Ti': 1, 'O', 3}
            elements: the dictionary keys for the composition
            elec_neg: the maximum pairwise electronegativity difference
            aos: the constituent atomic orbitals for each element as a dictionary
            band_edges: dictionary containing the highest occupied molecular orbital (HOMO),
                lowest unoccupied molecular orbital (LUMO), and whether the material is predicted
                to be a metal
        """
        self.composition = Composition(formula).as_dict()
        self.elements = list(self.composition)
        for subscript in self.composition.values():
            if not float(subscript).is_integer():
                raise ValueError("composition subscripts must be integers")

        self.elec_neg = self.max_electronegativity()
        self.aos = {str(el): [[str(el), k, v] for k, v in Element(el).atomic_orbitals.items()] for el in self.elements}
        self.band_edges = self.obtain_band_edges()

    def max_electronegativity(self) -> float:
        """
        Returns:
            The maximum pairwise electronegativity difference.
        """
        maximum: float = 0.0
        for e1, e2 in combinations(self.elements, 2):
            maximum = max(abs(Element(e1).X - Element(e2).X), maximum)
        return maximum

    def aos_as_list(self) -> list[tuple[str, str, float]]:
        """The orbitals energies in eV are represented as
        [['O', '1s', -18.758245], ['O', '2s', -0.871362], ['O', '2p', -0.338381]]
        Data is obtained from
        https://www.nist.gov/pml/data/atomic-reference-data-electronic-structure-calculations.

        Returns:
            A list of atomic orbitals, sorted from lowest to highest energy.
        """
        return sorted(
            chain.from_iterable([self.aos[el] * int(self.composition[el]) for el in self.elements]),  # type: ignore[misc]
            key=lambda x: x[2],
        )

    def obtain_band_edges(self) -> dict[str, Any]:
        """Fill up the atomic orbitals with available electrons.

        Returns:
            HOMO, LUMO, and whether it's a metal.
        """
        orbitals = self.aos_as_list()
        electrons = Composition(self.composition).total_electrons
        partial_filled = []
        for orbital in orbitals:
            if electrons <= 0:
                break
            if "s" in orbital[1]:
                electrons += -2
            elif "p" in orbital[1]:
                electrons += -6
            elif "d" in orbital[1]:
                electrons += -10
            elif "f" in orbital[1]:
                electrons += -14
            partial_filled.append(orbital)

        if electrons != 0:
            homo = partial_filled[-1]
            lumo = partial_filled[-1]
        else:
            homo = partial_filled[-1]
            try:
                lumo = orbitals[len(partial_filled)]
            except Exception:
                lumo = None

        return {"HOMO": homo, "LUMO": lumo, "metal": homo == lumo}
